# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

from enum import Enum
from abc import ABCMeta, abstractmethod

from btcpy.lib.types import Immutable, HexSerializable
from btcpy.structs.script import ScriptSig, StackData


class Branch(Enum):
    ELSE = 0
    IF = 1


class Sighash(Immutable, HexSerializable):

    types = {'ALL': 0x01,
             'NONE': 0x02,
             'SINGLE': 0x03}

    def __init__(self, sighash, anyonecanpay=False):
        if sighash not in self.__class__.types:
            raise ValueError('Unknown sighash: {}'.format(sighash))
        object.__setattr__(self, 'sighash', sighash)
        object.__setattr__(self, 'anyone', anyonecanpay)

    def __eq__(self, string):
        return self.sighash == string

    def __int__(self):
        return self.__class__.types[self.sighash] | (0x80 if self.anyone else 0x00)

    def __str__(self):
        return '[{}{}]'.format(self.sighash, '|ANYONECANPAY' if self.anyone else '')

    def as_byte(self):
        return int(self).to_bytes(1, 'little')

    def serialize(self):
        return int(self).to_bytes(4, 'little')


class Solver(metaclass=ABCMeta):
    
    @abstractmethod
    def solve(self, digest):
        raise NotImplemented

    @abstractmethod
    def get_sighashes(self):
        raise NotImplemented
    
    @abstractmethod
    def solves_segwit(self):
        raise NotImplemented
    
    def has_prev_script(self):
        return False
    
    
class SingleSigSolver(Solver, metaclass=ABCMeta):

    def __init__(self, sighash=Sighash('ALL')):
        self.sighash = sighash
    
    def get_sighashes(self):
        return [self.sighash]

    def solves_segwit(self):
        return False
    

class SegWitSolver(Solver, metaclass=ABCMeta):
    
    def solves_segwit(self):
        return True
    

class P2pkhSolver(SingleSigSolver):
    
    def __init__(self, privk, sighash=Sighash('ALL')):
        super().__init__(sighash)
        self.privk = privk
        
    def solve(self, digest):
        from .transaction import Witness
        pubkey = self.privk.pub()
        sig = self.privk.sign(digest)
        return (ScriptSig.from_stack_data([StackData.from_bytes(sig + self.sighash.as_byte()),
                                           StackData.from_bytes(pubkey.compressed)]),
                Witness([]))
        
        
class P2wpkhV0Solver(SegWitSolver, P2pkhSolver):
    
    def solve(self, digest):
        script_sig, witness = super().solve(digest)
        return witness.to_script_sig(), script_sig.to_witness()


class P2pkSolver(SingleSigSolver):
    
    def __init__(self, privk, sighash=Sighash('ALL')):
        super().__init__(sighash)
        self.privk = privk
        
    def solve(self, digest):
        from .transaction import Witness
        sig = self.privk.sign(digest)
        return (ScriptSig.from_stack_data([StackData.from_bytes(sig + self.sighash.as_byte())]),
                Witness([]))


class P2shSolver(Solver):
    
    def __init__(self, redeem_script, redeem_script_solver):
        self.redeem_script = redeem_script
        self.redeem_script_solver = redeem_script_solver
    
    def solve(self, *digests):
        script_sig, witness = self.redeem_script_solver.solve(*digests)
        script_sig_data = script_sig.get_data()
        script_sig_data.append(self.redeem_script.to_stack_data())
        return ScriptSig.from_stack_data(script_sig_data), witness
    
    def get_sighashes(self):
        return self.redeem_script_solver.get_sighashes()

    def solves_segwit(self):
        return self.redeem_script_solver.solves_segwit()
    
    def get_prev_script(self):
        if self.redeem_script_solver.has_prev_script():
            return self.redeem_script_solver.get_prev_script()
        else:
            return self.redeem_script
    
    def has_prev_script(self):
        return True


class P2wshV0Solver(SegWitSolver):
    
    def __init__(self, witness_script, witness_script_solver):
        self.witness_script = witness_script
        self.witness_script_solver = witness_script_solver

    def solve(self, *digests):
        from .transaction import Witness
        script_sig, witness = self.witness_script_solver.solve(*digests)
        return (ScriptSig.empty(),
                (script_sig.to_witness()
                 + witness
                 + Witness([self.get_prev_script().to_stack_data()])))
    
    def get_sighashes(self):
        return self.witness_script_solver.get_sighashes()
    
    def get_prev_script(self):
        return self.witness_script
    
    def has_prev_script(self):
        return True


class MultisigSolver(Solver):
    
    def __init__(self, *privkeys, sighashes=None):
        if sighashes is None:
            sighashes = [Sighash('ALL') for _ in privkeys]
        if len(sighashes) != len(privkeys):
            raise ValueError('{} privkeys provided and {} sighashes'.format(len(privkeys), len(sighashes)))
        self.privkeys = privkeys
        self.sighashes = sighashes
        
    def solve(self, *digests):
        from .transaction import Witness
        if len(digests) != len(self.privkeys):
            raise ValueError('{} privkeys provided and {} digests'.format(len(self.privkeys), len(digests)))
        script_sig_data = [StackData.zero()]
        for priv, digest, sighash in zip(self.privkeys, digests, self.sighashes):
            sig = priv.sign(digest)
            script_sig_data.append(StackData.from_bytes(sig + sighash.as_byte()))
        return ScriptSig.from_stack_data(script_sig_data), Witness([])
    
    def get_sighashes(self):
        return self.sighashes

    def solves_segwit(self):
        return False
        

class IfElseSolver(Solver):
    
    def __init__(self, branch, inner_solver):
        self.branch = branch
        self.inner_solver = inner_solver
        
    def solve(self, *digests):
        script_sig, witness = self.inner_solver.solve(*digests)
        script_sig_data = script_sig.get_data()
        script_sig_data.append(StackData.from_int(self.branch.value))
        return ScriptSig.from_stack_data(script_sig_data), witness
    
    def get_sighashes(self):
        return self.inner_solver.get_sighashes()

    def solves_segwit(self):
        return self.inner_solver.solves_segwit()


class TimelockSolver(Solver):
    
    def __init__(self, inner_solver):
        self.inner_solver = inner_solver
        
    def solve(self, *digests):
        return self.inner_solver.solve(*digests)
    
    def get_sighashes(self):
        return self.inner_solver.get_sighashes()

    def solves_segwit(self):
        return self.inner_solver.solves_segwit()
    
    
class HashlockSolver(Solver):
    
    def __init__(self, preimage, inner_solver):
        self.preimage = preimage
        self.inner_solver = inner_solver
        
    def solve(self, *digests):
        script_sig, witness = self.inner_solver.solve(*digests)
        script_sig_data = script_sig.get_data()
        script_sig_data.append(StackData.from_bytes(self.preimage))
        return ScriptSig.from_stack_data(script_sig_data), witness
        
    def get_sighashes(self):
        return self.inner_solver.get_sighashes()

    def solves_segwit(self):
        return self.inner_solver.solves_segwit()



from collections import namedtuple
from decimal import Decimal


Constants = namedtuple('Constants', [
    'name',
    'base58_prefixes',
    'base58_raw_prefixes',
    'bech32_hrp',
    'bech32_net',
    'xkeys_prefix',
    'xpub_version',
    'xprv_version',
    'wif_prefix',
    'from_unit',
    'to_unit'
])


BitcoinMainnet = Constants(
    name='BitcoinMainnet',
    base58_prefixes={
        '1': 'p2pkh',
        '3': 'p2sh',
    },
    base58_raw_prefixes={
        'p2pkh': bytearray(b'\x00'),
        'p2sh': bytearray(b'\x05'),
    },
    bech32_hrp='bc',
    bech32_net='mainnet',
    xkeys_prefix='x',
    xpub_version=b'\x04\x88\xb2\x1e',
    xprv_version=b'\x04\x88\xad\xe4',
    wif_prefix=0x80,
    from_unit=Decimal('1e-8'),
    to_unit=Decimal('1e8')
)


BitcoinTestnet = Constants(
    name='BitcoinTestnet',
    base58_prefixes={
        'm': 'p2pkh',
        'n': 'p2pkh',
        '2': 'p2sh',
    },
    base58_raw_prefixes={
        'p2pkh': bytearray(b'\x6f'),
        'p2sh': bytearray(b'\xc4'),
    },
    bech32_hrp='tb',
    bech32_net='testnet',
    xkeys_prefix='t',
    xpub_version=b'\x04\x35\x87\xcf',
    xprv_version=b'\x04\x35\x83\x94',
    wif_prefix=0xef,
    from_unit=Decimal('1e-8'),
    to_unit=Decimal('1e8')
)


PeercoinMainnet = Constants(
    name='PeercoinMainnet',
    base58_prefixes={
        'P': 'p2pkh',
        'p': 'p2sh',
    },
    base58_raw_prefixes={
        'p2pkh': bytearray(b'\x37'),
        'p2sh': bytearray(b'\x75'),
    },
    bech32_hrp='bc',
    bech32_net='mainnet',
    xkeys_prefix='x',
    xpub_version=b'\x04\x88\xb2\x1e',
    xprv_version=b'\x04\x88\xad\xe4',
    wif_prefix=0xb7,
    from_unit=Decimal('1e-6'),
    to_unit=Decimal('1e6')
)


PeercoinTestnet = Constants(
    name='PeercoinTestnet',
    base58_prefixes={
        'm': 'p2pkh',
        'n': 'p2pkh',
    },
    base58_raw_prefixes={
        'p2pkh': bytearray(b'\x6f'),
        'p2sh': bytearray(b'\xc4'),
    },
    bech32_hrp='tb',
    bech32_net='testnet',
    xkeys_prefix='t',
    xpub_version=b'\x04\x35\x87\xcf',
    xprv_version=b'\x04\x35\x83\x94',
    wif_prefix=0xef,
    from_unit=Decimal('1e-6'),
    to_unit=Decimal('1e6')
)



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

import hmac
from hashlib import sha512
from ..lib.base58 import b58decode_check, b58encode_check
from ecdsa import VerifyingKey
from ecdsa.ellipticcurve import INFINITY
from ecdsa.curves import SECP256k1
from ecdsa.ecdsa import generator_secp256k1
from abc import ABCMeta, abstractmethod

from ..lib.types import HexSerializable
from ..lib.parsing import Stream, Parser
from .crypto import PrivateKey, PublicKey


class ExtendedKey(HexSerializable, metaclass=ABCMeta):

    master_parent_fingerprint = bytearray([0]*4)
    first_hardened_index = 1 << 31
    curve_order = SECP256k1.order

    @classmethod
    def master(cls, key, chaincode):
        return cls(key, chaincode, 0, cls.master_parent_fingerprint, 0, hardened=True)

    @classmethod
    def decode(cls, network, string):

        if string[0] != network.xkeys_prefix:
            raise ValueError('Encoded key not recognised: {}'.format(string))

        decoded = b58decode_check(string)
        parser = Parser(bytearray(decoded))
        parser >> 4
        depth = int.from_bytes(parser >> 1, 'big')
        fingerprint = parser >> 4
        index = int.from_bytes(parser >> 4, 'big')

        if index >= cls.first_hardened_index:
            index -= cls.first_hardened_index
            hardened = True
        else:
            hardened = False

        chaincode = parser >> 32
        keydata = parser >> 33

        if string[1:4] == 'prv':
            subclass = ExtendedPrivateKey
        elif string[1:4] == 'pub':
            subclass = ExtendedPublicKey
        else:
            raise ValueError('Encoded key not recognised: {}'.format(string))

        key = subclass.decode_key(keydata)

        return subclass(key, chaincode, depth, fingerprint, index, hardened)

    @staticmethod
    @abstractmethod
    def decode_key(keydata):
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def get_version():
        raise NotImplemented

    def __init__(self, key, chaincode, depth, pfing, index, hardened=False):
        if not 0 <= depth <= 255:
            raise ValueError('Depth must be between 0 and 255')
        self.key = key
        self.chaincode = chaincode
        self.depth = depth
        self.parent_fingerprint = pfing
        self.index = index
        self.hardened = hardened

    def derive(self, path):
        """
        :param path: a path like "m/44'/0'/1'/0/10" if deriving from a master key,
                     or a relative path like "./0/10"
        :return: the derived ExtendedPublicKey if deriving from an ExtendedPublicKey,
                 the derived ExtendedPrivateKey if deriving from an ExtendedPrivateKey
        """
        steps = path.split('/')

        if steps[0] not in {'m', '.'}:
            raise ValueError('Invalid derivation path: {}'.format(path))

        if steps[0] == 'm' and not self.is_master():
            raise ValueError('Trying to derive absolute path from non-master key')

        current = self
        for step in steps[1:]:
            hardened = False
            if step[-1] == "'":
                hardened = True
                step = step[:-1]
            index = int(step)
            current = current.get_child(index, hardened)

        return current

    @abstractmethod
    def get_child(self, index, hardened=False):
        raise NotImplemented

    @abstractmethod
    def _serialized_public(self):
        raise NotImplemented

    @abstractmethod
    def _serialize_key(self):
        raise NotImplemented

    def get_hash(self, index, hardened=False):
        cls = self.__class__
        if hardened:
            data = self._serialize_key() + (index + cls.first_hardened_index).to_bytes(4, 'big')
        else:
            data = self._serialized_public() + index.to_bytes(4, 'big')
        h = bytearray(hmac.new(self.chaincode, data, sha512).digest())
        left, right = int.from_bytes(h[:32], 'big'), h[32:]
        if left > cls.curve_order:
            raise ValueError('Left side of hmac generated number bigger than SECP256k1 curve order')
        return left, right

    def is_master(self):
        return all([self.depth == 0,
                    self.parent_fingerprint == ExtendedKey.master_parent_fingerprint,
                    self.index == 0])

    def encode(self, network):
        return b58encode_check(bytes(self.serialize(network)))

    def serialize(self, network):
        cls = self.__class__
        result = Stream()
        result << cls.get_version(network)
        result << self.depth.to_bytes(1, 'big')
        result << self.parent_fingerprint
        if self.hardened:
            result << (self.index + cls.first_hardened_index).to_bytes(4, 'big')
        else:
            result << self.index.to_bytes(4, 'big')
        result << self.chaincode
        result << self._serialize_key()
        return result.serialize()

    def __str__(self):
        return 'version: {}\ndepth: {}\nparent fp: {}\n' \
               'index: {}\nchaincode: {}\nkey: {}\nhardened: {}'.format(self.__class__.get_version(),
                                                                        self.depth,
                                                                        self.parent_fingerprint,
                                                                        self.index,
                                                                        self.chaincode,
                                                                        self.key,
                                                                        self.hardened)

    def __eq__(self, other):
        return all([self.key == other.key,
                    self.chaincode == other.chaincode,
                    self.depth == other.depth,
                    self.parent_fingerprint == other.parent_fingerprint,
                    self.index == other.index,
                    self.hardened == other.hardened])


class ExtendedPrivateKey(ExtendedKey):

    @staticmethod
    def get_version(network):
        return network.xprv_version

    @staticmethod
    def decode_key(keydata):
        return PrivateKey(keydata[1:])

    def __init__(self, key, chaincode, depth, pfing, index, hardened=False):
        if not isinstance(key, PrivateKey):
            raise TypeError('ExtendedPrivateKey expects a PrivateKey')
        super().__init__(key, chaincode, depth, pfing, index, hardened)

    def __int__(self):
        return int.from_bytes(self.key.key, 'big')

    def get_child(self, index, hardened=False):
        left, right = self.get_hash(index, hardened)
        k = (int(self) + left) % self.__class__.curve_order
        if k == 0:
            raise ValueError('Got 0 as k')
        return ExtendedPrivateKey(PrivateKey(k.to_bytes(32, 'big')),
                                  right,
                                  self.depth + 1,
                                  self.get_fingerprint(),
                                  index,
                                  hardened)

    def get_fingerprint(self):
        return self.pub().get_fingerprint()

    def _serialize_key(self):
        return bytearray([0]) + self.key.serialize()

    def _serialized_public(self):
        return self.pub()._serialize_key()

    def pub(self):
        return ExtendedPublicKey(self.key.pub(),
                                 self.chaincode,
                                 self.depth,
                                 self.parent_fingerprint,
                                 self.index,
                                 self.hardened)


class ExtendedPublicKey(ExtendedKey):

    @staticmethod
    def get_version(network):
        return network.xpub_version

    @staticmethod
    def decode_key(keydata):
        return PublicKey(keydata)

    def __init__(self, key, chaincode, depth, pfing, index, hardened=False):
        if not isinstance(key, PublicKey):
            raise TypeError('ExtendedPublicKey expects a PublicKey')
        super().__init__(key.compress(), chaincode, depth, pfing, index, hardened)

    def __int__(self):
        return int.from_bytes(self.key.key, 'big')

    def get_fingerprint(self):
        return self.key.hash()[:4]

    def get_child(self, index, hardened=False):
        left, right = self.get_hash(index, hardened)
        point = ((left * generator_secp256k1)
                 + VerifyingKey.from_string(self.key.uncompressed[1:], curve=SECP256k1).pubkey.point)
        if point == INFINITY:
            raise ValueError('Computed point equals INFINITY')
        return ExtendedPublicKey(PublicKey.from_point(point), right, self.depth+1, self.get_fingerprint(), index, False)

    def get_hash(self, index, hardened=False):
        if hardened:
            raise ValueError('Trying to generate hardened child from public key')
        return super().get_hash(index, hardened)

    def _serialized_public(self):
        return self._serialize_key()

    def _serialize_key(self):
        return self.key.compressed

    def __lt__(self, other):
        return self.key < other.key



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

from binascii import hexlify, unhexlify
from hashlib import sha256
import hashlib

from .types import Serializable, HexSerializable
from .opcodes import OpCodeConverter


class IncompleteParsingException(ValueError):
    pass


class Parser(object):
    @staticmethod
    def to_varint(integer):
        if 0 <= integer < 0xfd:
            prefix, size = bytearray(), 1
        elif 0xfd <= integer <= 0xffff:
            prefix, size = bytearray([0xfd]), 2
        elif 0xffff <= integer <= 0xffffffff:
            prefix, size = bytearray([0xfe]), 4
        elif 0xffffffff <= integer <= 0xffffffffffffffff:
            prefix, size = bytearray([0xff]), 8
        else:
            raise ValueError('Wrong value for varint: {}'.format(integer))
        return prefix + bytearray(integer.to_bytes(size, 'little'))

    @staticmethod
    def from_hex(hexa):
        return Parser(unhexlify(hexa))

    def __init__(self, bytes_):
        self._string = bytes_
        self.pointer = 0

    def __rshift__(self, bytes_) -> bytearray:
        """
        Moves the parser forward of `bytes_` bytes. This consumes the first
        `bytes_` bytes of the string, and returns them
        :param bytes_: number of bytes to shift the pointer
        :return: part of the string consumed
        """
        if self.pointer == len(self._string):
            raise StopIteration('Trying to shift empty string')

        result = self._string[self.pointer:self.pointer + bytes_]
        if len(result) != bytes_:
            raise StopIteration('Not enough bytes available')

        self.pointer += bytes_

        return result

    def __lshift__(self, bytes_) -> bytearray:
        if len(self._string) == 0:
            raise StopIteration("Trying to shift the empty string")
        if self.pointer - bytes_ < 0:
            raise StopIteration("Trying to lshift before the beginning of the string")
        self.pointer -= bytes_

        result = self._string[self.pointer:self.pointer + bytes_]
        return result

    def __bool__(self):
        return self.pointer < len(self._string)

    def __next__(self):
        return (self >> 1)[0]

    def __iter__(self):
        return self

    def __len__(self):
        return len(self._string) - self.pointer

    def __getitem__(self, item):
        return self._string[item + self.pointer]

    def parse_varint(self):
        header = next(self)
        if header == 0xfd:
            result = self >> 2
        elif header == 0xfe:
            result = self >> 4
        elif header == 0xff:
            result = self >> 8
        else:
            return header
        return int.from_bytes(result, 'little')


class BlockHeaderParser(Parser):
    def __init__(self, bytes_):
        super().__init__(bytes_)

    def get_version(self):
        return int.from_bytes(self >> 4, 'little')

    def get_prev_block_hash(self):
        return hexlify((self >> 32)[::-1]).decode()

    def get_merkle_root(self):
        return hexlify((self >> 32)[::-1]).decode()

    def get_block_timestamp(self):
        return int.from_bytes(self >> 4, 'little')

    def get_bits(self):
        return hexlify((self >> 4)[::-1]).decode()

    def get_block_nonce(self):
        return int.from_bytes(self >> 4, 'little')

    def get_block_header(self):
        from ..structs.block import BlockHeader
        version = self.get_version()
        prev_block_hash = self.get_prev_block_hash()
        merkle_root = self.get_merkle_root()
        block_timestamp = self.get_block_timestamp()
        bits = self.get_bits()
        block_nonce = self.get_block_nonce()
        return BlockHeader(version, prev_block_hash, merkle_root, block_timestamp, bits, block_nonce)


class BlockParser(BlockHeaderParser):

    def __init__(self, bytes_, tx_parser, network):
        super().__init__(bytes_)
        self.tx_parser = tx_parser
        self.network = network

    def get_txn_count(self):

        return self.parse_varint()

    def get_txns(self):

        txn_count = self.get_txn_count()
        counter = 0
        txns_parser = self.tx_parser(self >> len(self), self.network)
        txns = []
        for i in range(txn_count):
            txns.append(txns_parser.get_next_tx())
            counter += 1
        if len(txns_parser) != 0:
            raise IncompleteParsingException("Incomplete Block parsing, leftover data...")
        return txns


class TransactionParser(Parser):

    def __init__(self, bytes_, network):
        super().__init__(bytes_)
        self.network = network
        self.segwit = False
        self.txins = 0

    def _version(self):
        return int.from_bytes(self >> 4, 'little')

    def _txin_data(self):
        from ..structs.script import ScriptSig, CoinBaseScriptSig
        from ..structs.transaction import Sequence
        txout_hash = hexlify((self >> 32)[::-1]).decode()
        txout_index = int.from_bytes(self >> 4, 'little')

        if txout_hash == '0' * 64 and txout_index == 0xffffffff:
            script = CoinBaseScriptSig(self >> self.parse_varint())
        else:
            script = ScriptSig(self >> self.parse_varint())
        sequence = Sequence(int.from_bytes(self >> 4, 'little'))
        return txout_hash, txout_index, script, sequence

    def _txins_data(self):
        """
        :return: a couple (segwit, txins) where `segwit` is a boolean
        telling whether the transaction has a segwit format
        """
        ntxins = self.parse_varint()
        if ntxins == 0:
            self.segwit = True
            flag = next(self)
            if flag != 1:
                raise ValueError('Wrong flag in SegWit transaction: {}'.format(flag))
            ntxins = self.parse_varint()
        else:
            self.segwit = False
        self.txins = ntxins
        return self.segwit, [self._txin_data() for _ in range(ntxins)]

    def _txout(self, n):
        from ..structs.script import ScriptBuilder
        from ..structs.transaction import TxOut

        value = int.from_bytes(self >> 8, 'little')
        script = ScriptBuilder.identify(self >> self.parse_varint())
        return TxOut(value, n, script, self.network)

    def _txouts(self):
        return [self._txout(i) for i in range(self.parse_varint())]

    def _witness(self):
        from ..structs.script import StackData
        if self.segwit:
            witnesses = []
            for _ in range(self.txins):
                witnesses.append(
                    [StackData.from_bytes(self >> self.parse_varint()) for _ in range(self.parse_varint())])
            return witnesses
        raise ValueError('Trying to get witness on a non-segwit transaction')

    def _locktime(self):
        from ..structs.transaction import Locktime
        return Locktime(int.from_bytes(self >> 4, 'little'))

    def get_next_tx(self, mutable=False):
        from ..structs.script import CoinBaseScriptSig
        from ..structs.transaction import (CoinBaseTxIn, Witness, TxIn, SegWitTransaction, Transaction)

        version = self._version()
        segwit, txins_data = self._txins_data()
        txouts = self._txouts()
        if segwit:
            witness = self._witness()
            txins = [CoinBaseTxIn(*txin_data[2:], witness=Witness(wit))
                     if isinstance(txin_data[2], CoinBaseScriptSig)
                     else TxIn(*txin_data, witness=Witness(wit))
                     for txin_data, wit in zip(txins_data, witness)]
        else:
            txins = [CoinBaseTxIn(*txin_data[2:])
                     if isinstance(txin_data[2], CoinBaseScriptSig)
                     else TxIn(*txin_data)
                     for txin_data in txins_data]

        locktime = self._locktime()

        if len(txins) > 1 and isinstance(txins[0], CoinBaseTxIn):
            raise ValueError('Transaction looks like coinbase but has more than one txin')

        if segwit:
            result = SegWitTransaction(version, txins, txouts, locktime, self.network)
        else:
            result = Transaction(version, txins, txouts, locktime, self.network)

        return result.to_mutable() if mutable else result


class PeercoinTxParser(TransactionParser):

    def _timestamp(self):
        '''get transaction timestamp (peercoin specific)'''
        return int.from_bytes(self >> 4, 'little')

    def _txout(self, n):
        from ..structs.script import ScriptBuilder
        from ..structs.transaction import PeercoinTxOut

        value = int.from_bytes(self >> 8, 'little')
        script = ScriptBuilder.identify(self >> self.parse_varint())
        return PeercoinTxOut(value, n, script, self.network)

    def get_next_tx(self, mutable=False):
        from ..structs.script import CoinBaseScriptSig
        from ..structs.transaction import (CoinBaseTxIn, Witness, TxIn, PeercoinTx)

        version = self._version()
        tstamp = self._timestamp()
        segwit, txins_data = self._txins_data()
        txouts = self._txouts()
        if segwit:
            witness = self._witness()
            txins = [CoinBaseTxIn(*txin_data[2:], witness=Witness(wit))
                     if isinstance(txin_data[2], CoinBaseScriptSig)
                     else TxIn(*txin_data, witness=Witness(wit))
                     for txin_data, wit in zip(txins_data, witness)]
        else:
            txins = [CoinBaseTxIn(*txin_data[2:])
                     if isinstance(txin_data[2], CoinBaseScriptSig)
                     else TxIn(*txin_data)
                     for txin_data in txins_data]

        locktime = self._locktime()

        if len(txins) > 1 and isinstance(txins[0], CoinBaseTxIn):
            raise ValueError('Transaction looks like coinbase but has more than one txin')

        if segwit:
            raise Exception('Peercoin does not currently support SegWit.')
        else:
            result = PeercoinTx(version, tstamp, txins, txouts, locktime)

        return result.to_mutable() if mutable else result


class UnexpectedOperationFound(Exception):
    pass


class ScriptParser(Parser):
    def match(self, template, end=True):
        ops = iter(template.split(' '))
        pushes = []
        for op in ops:
            if OpCodeConverter.exists(op):
                self.require(op)
            elif '<' in op and '>' in op:  # push operation
                if op[-1] == '*':
                    pushes += self.require_pushes(zero=True)
                elif op[-1] == '+':
                    pushes += self.require_pushes(zero=False)
                elif op[-1] == '>':
                    pushes.append(self.require_push(op[1:-1]))
                else:
                    raise ValueError('Could not parse push requirement: {}'.format(op))
            else:
                raise ValueError('Could not understand template')
        if end:
            self.require_empty()
        return pushes

    def push_until(self, op, zero=False):
        pushes = self.require_pushes(zero)
        self.require(op)
        return pushes

    def require_pushes(self, zero=False):
        pushes = []
        try:
            while True:
                next_op = next(self)
                # print('Parsing op: {}'.format(next_op))
                pushes.append(self.get_push(next_op))
        except UnexpectedOperationFound:
            # could not push, probably not a push operation, let's restore the buffer to the previous character
            self << 1
        except StopIteration:
            # we reached the end of the buffer
            pass
        finally:
            if not zero and len(pushes) == 0:
                raise UnexpectedOperationFound('Found zero pushes, though more than zero required')
            return pushes

    def require_push(self, constraint):
        try:
            push_data = self.get_push()
        except StopIteration:
            raise UnexpectedOperationFound('Empty push where push was required')
        validator = PushValidator(constraint)
        if not validator.check(push_data):
            raise UnexpectedOperationFound('Push operation did not pass validation constraints: {} '
                                           'with data: {}'.format(constraint, push_data.hexlify()))
        return push_data

    def require(self, op):
        try:
            next_op = next(self)
        except StopIteration:
            raise UnexpectedOperationFound('No further operation: parser reached end of script')
        if next_op != OpCodeConverter.to_int(op):
            raise UnexpectedOperationFound('{} not found, found {} instead'.format(op, next_op))

    def require_empty(self):
        if self:
            raise UnexpectedOperationFound('Unexpected operations left after template match')

    def get_push(self, curr_op=None):
        from ..structs.script import StackData
        try:
            if curr_op is None:
                curr_op = next(self)
        except StopIteration:
            raise
        return StackData.from_push_op(self, curr_op)


class PushValidator(object):
    sep = '|'
    range_sep = '-'

    def check(self, pushdata):
        for const in self.constraints:
            if PushValidator.range_sep in const:
                minimum, maximum = const.split(PushValidator.range_sep)
                if len(pushdata) in range(int(minimum), int(maximum) + 1):
                    return True
            elif not const or len(pushdata) == int(const):
                return True
        return False

    def __init__(self, string):
        if PushValidator.sep in string:
            self.constraints = string.split(PushValidator.sep)
        else:
            self.constraints = [string]


class Stream(HexSerializable):
    @staticmethod
    def unhexlify(hex_string):
        return Stream(bytearray(unhexlify(hex_string)))

    def __init__(self, initial=None):
        if initial is None:
            initial = bytearray()
        self.body = bytearray()
        self << initial

    def __lshift__(self, other):
        if isinstance(other, Serializable):
            self.body += other.serialize()
        elif isinstance(other, (bytes, bytearray)):
            self.body += other
        elif isinstance(other, int):
            self.body.append(other)
        else:
            raise ValueError('Cannot insert {} in stream'.format(type(other)))
        return self

    def __add__(self, other):
        new = Stream(self.body)
        new << other
        return new

    def __bool__(self):
        return bool(self.body)

    def sha256(self):
        return bytearray(sha256(self.body).digest())

    def ripemd(self):
        ripe = hashlib.new('ripemd160')
        ripe.update(self.body)
        return bytearray(ripe.digest())

    def hash256(self):
        return bytearray(sha256(sha256(self.body).digest()).digest())

    def hash160(self):
        sha = sha256(self.body).digest()
        ripe = hashlib.new('ripemd160')
        ripe.update(sha)
        return bytearray(ripe.digest())

    def serialize(self):
        return self.body



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.




# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

import json
import hashlib
from hashlib import sha256
from binascii import hexlify, unhexlify
from abc import ABCMeta, abstractmethod

from ..lib.types import HexSerializable, Immutable, cached
from ..lib.parsing import ScriptParser, Parser, Stream, UnexpectedOperationFound
from ..lib.opcodes import OpCodeConverter
from .crypto import WrongPubKeyFormat, PublicKey
from .address import P2pkhAddress, P2shAddress, P2wpkhAddress, P2wshAddress
from btcpy.constants import BitcoinMainnet


class WrongScriptTypeException(Exception):
    pass


class WrongPushDataOp(Exception):
    pass


# noinspection PyUnresolvedReferences
class StackData(Immutable, HexSerializable):
    """
    Represents data pushed on the stack. This class has commodity methods to get the length in bytes of the data
    and the push operation needed in a script to push those data on the stack
    """

    max_size = 2**32

    @classmethod
    def zero(cls):
        return cls(bytearray([0]))

    @classmethod
    def unhexlify(cls, hexa):
        return cls.from_bytes(bytearray(unhexlify(hexa)))

    @classmethod
    def from_bytes(cls, bytes_):
        data_len = len(bytes_)
        if data_len == 0:
            return cls.zero()

        if data_len == 1:
            if 1 <= bytes_[0] <= 16:
                return cls(bytearray([80 + bytes_[0]]))

        # this should execute when data_len == 1 and it
        # does not fall in the previous conditions
        if data_len <= 75:
            return cls(bytearray([data_len]), bytes_)
        else:
            if data_len <= 0xff:
                size = 1
            elif data_len <= 0xffff:
                size = 2
            elif data_len <= 0xffffffff:
                size = 4
            else:
                raise ValueError('Data length too big to push: {} bytes'.format(data_len))
            return cls(bytearray([OpCodeConverter.to_int('OP_PUSHDATA{}'.format(size))])
                       + bytearray(data_len.to_bytes(size, 'little')),
                       bytes_)

    @classmethod
    def from_int(cls, integer):
        from math import ceil
        if integer == 0:
            return cls.zero()
        if integer == -1:
            return cls(bytearray([0x4f]))
        sign = True if integer < 0 else False
        absolute = bytearray(abs(integer).to_bytes(ceil(abs(integer).bit_length() / 8), 'little'))
        if absolute[-1] & (1 << 7):
            # using all the bits: need to add a whole byte for the sign
            absolute.append((1 << 7) if sign else 0)
        else:
            absolute[-1] |= (1 << 7) if sign else 0
        return cls.from_bytes(absolute)

    @classmethod
    def from_push_op(cls, parser, push_op):
        try:

            if push_op == 0:
                return cls.zero()

            if 1 <= push_op <= 75:
                return cls(bytearray([push_op]), parser >> push_op)

            if 76 <= push_op <= 78:
                push_size_in_bytes = parser >> 2**(push_op - 76)  # 1, 2 or 4 depending on the push op
                push_size = int.from_bytes(push_size_in_bytes, 'little')
                return cls(bytearray([push_op]) + push_size_in_bytes, parser >> push_size)

            if 79 <= push_op <= 96:
                if push_op == 80:
                    raise UnexpectedOperationFound('Trying to use OP_RESERVED (0x50) as a push operation')
                return cls(bytearray([push_op]))

            raise UnexpectedOperationFound('Invalid push operation: {}'.format(push_op))

        except StopIteration:
            raise WrongPushDataOp('Reached end of script while parsing pushdata operation')

    @staticmethod
    def check_op(push_op, data):

        if len(data) > StackData.max_size:
            raise WrongPushDataOp('Trying to create StackData with size: {}'.format(len(data)))

        if not push_op:
            raise WrongPushDataOp('Empty bytearray is not a valid push operation')

        if len(push_op) == 1:
            # one-byte push op
            if 1 <= push_op[0] <= 75:
                # push op must equal len(data)
                if push_op[0] != len(data):
                    raise WrongPushDataOp('Push op does not match data length: {}, {}'.format(push_op, len(data)))
            elif 79 <= push_op[0] <= 96 or push_op[0] == 0:
                # TODO understand whether this breaks consensus rules!
                if push_op[0] == 80:
                    raise WrongPushDataOp('Trying to use OP_RESERVED (0x50) as a push operation')
                # pushing data between 0 and 16, no further data required
                if data:
                    raise WrongPushDataOp('Push op is OP_{} but some data was provided: {}'
                                          .format(0 if push_op[0] == 0 else push_op[0] - 80, data))
            else:
                raise WrongPushDataOp('One-byte push operation not recognized: {}'.format(push_op[0]))
        elif 76 <= push_op[0] <= 78:
            n = 1 + 2**(push_op[0] - 76)  # number of bytes required for the push op
            if len(push_op) != n:
                raise WrongPushDataOp('We need {} bytes for PUSHDATA{}, found {}'.format(n, n-1, len(push_op)))
            if int.from_bytes(push_op[1:n], 'little') != len(data):
                raise WrongPushDataOp('Push op does not match data length: {}, {}'.format(push_op, len(data)))
        else:
            raise WrongPushDataOp('Push operation not recognized: {}'.format(push_op))

    def __init__(self, push_op, data=None):
        """
        :param push_op: a bytearray representing a push operation
        :param data: a bytearray of stack data
        """
        if data is None:
            data = bytearray()

        StackData.check_op(push_op, data)

        object.__setattr__(self, 'push_op', push_op)
        object.__setattr__(self, 'data', bytearray(data))

    def __str__(self):
        if self.push_op[0] == 0 or 81 <= self.push_op[0] <= 96:
            if self.push_op[0] == 0:
                return ''
            else:
                return OpCodeConverter.from_int(self.push_op[0])
        if not self.data:
            return '00'
        return hexlify(self.data).decode()

    def __len__(self):
        if self.push_op[0] == 0 or 81 <= self.push_op[0] <= 96:
            # OP_0 to OP_16
            return 1
        return len(self.data)

    def __bool__(self):
        return len(self) != 0

    def __int__(self):

        if self.push_op[0] == 0:
            return 0

        if 79 <= self.push_op[0] <= 96:
            return self.push_op[0] - 80

        sign = bool(self.data[-1] & (1 << 7))  # store sign
        first_byte = self.data[-1] & ~(1 << 7)  # remove sign
        absolute = int.from_bytes(self.data[:-1] + bytearray([first_byte]), 'little')
        return absolute * (-1 if sign else 1)

    def to_push_op(self):
        return self.push_op + self.data

    @cached
    def serialize(self):
        if self.push_op[0] == 0:
            return Parser.to_varint(0) + bytearray()

        if self.push_op[0] in (79, 80):
            return Parser.to_varint(1) + bytearray([self.push_op[0]])

        if self.push_op[0] in range(81, 97):
            return Parser.to_varint(1) + bytearray([self.push_op[0] - 80])

        # 1-75
        return Parser.to_varint(len(self)) + self.data


# noinspection PyUnresolvedReferences
class BaseScript(Immutable, HexSerializable, metaclass=ABCMeta):

    # noinspection PyMethodOverriding
    @classmethod
    def from_json(cls, string):
        dic = json.loads(string)
        return cls(dic['hex'])

    @classmethod
    def unhexlify(cls, hex_string):
        return cls(bytearray(unhexlify(hex_string)))

    @staticmethod
    def compile(string):
        result = Stream()
        opcodes = string.split(' ')
        for opcode in opcodes:
            try:
                result << OpCodeConverter.to_int(opcode)
            except ValueError:
                result << StackData.unhexlify(opcode).to_push_op()
        return result.serialize()

    def __init__(self, bytes_, strict=True):
        object.__setattr__(self, 'strict', strict)
        object.__setattr__(self, 'body', bytes_)

    def __len__(self):
        return len(self.body)

    def __str__(self):
        return self.decompile()

    def __add__(self, other):
        return UnknownScript(self.body + other.body)

    def __eq__(self, other):
        return self.body == other.body

    def __iter__(self):
        return iter(self.decompile().split())

    def serialize(self):
        return self.body

    @cached
    def decompile(self):
        opcodes = []

        parser = ScriptParser(self.body)

        while parser:
            op = next(parser)
            if 1 <= op <= 78:  # pushdata
                try:
                    pushed_data = parser.get_push(op)
                except WrongPushDataOp:
                    opcodes.append('[error]')
                    break
                opcodes.append(pushed_data)
            else:
                try:
                    opcodes.append(OpCodeConverter.from_int(op))
                except ValueError:
                    raise ValueError('Invalid opcode {} in script {}'.format(op, self.hexlify()))

        return ' '.join(str(opcode) for opcode in opcodes)

    def is_standard(self):
        pass

    @cached
    def get_sigop_count(self):
        sigops = 0
        lastop = None
        for op in self:
            if op in {'OP_CHECKSIG', 'OP_CHECKSIGVERIFY'}:
                sigops += 1
            elif op in {'OP_CHECKMULTISIG', 'OP_CHECKMULTISIGVERIFY'}:
                if lastop is None:
                    # OP_CHECKMULTISIG(VERIFY) is the first op of the script. This means the script is invalid
                    return 0
                if OpCodeConverter.to_int('OP_1') <= OpCodeConverter.to_int(lastop) <= OpCodeConverter.to_int('OP_16'):
                    sigops += OpCodeConverter.to_int(lastop) - 80
            lastop = op
        return sigops

    @property
    @abstractmethod
    def type(self):
        raise NotImplemented

    @cached
    def is_push_only(self):
        for op in self:
            try:
                if OpCodeConverter.to_int(op) > OpCodeConverter.to_int('OP_16'):
                    return False
            except ValueError:
                continue
        return True

    @cached
    def get_data(self):
        parser = ScriptParser(self.body)
        data = []
        while True:
            try:
                data.append(parser.get_push())
            except UnexpectedOperationFound:
                pass
            except StopIteration:
                break
        return data


class Script(BaseScript):

    @property
    def type(self):
        return 'Script'


class ScriptSig(BaseScript):

    @staticmethod
    def empty():
        return ScriptSig(bytearray())

    @staticmethod
    def from_stack_data(stack_data):
        return ScriptSig(bytearray([byte_ for item in stack_data for byte_ in item.to_push_op()]))

    @cached
    def to_json(self):
        return {'asm': str(self),
                'hex': self.hexlify()}

    def is_standard(self, spends_p2sh=False):

        if len(self) > 1650:
            return False

        if not self.is_push_only():
            return False

        if spends_p2sh:
            redeem_script = ScriptBuilder.identify(self.get_data()[-1].data)
            if redeem_script.get_sigop_count() > 15:
                return False

        return True

    @cached
    def to_witness(self):
        from .transaction import Witness
        return Witness(self.get_data())

    @property
    def type(self):
        return 'scriptSig'


class CoinBaseScriptSig(ScriptSig):

    def is_standard(self, *args, **kwargs):
        if len(self) > 1650:
            return False
        return True

    def to_json(self):
        return {'hex': self.hexlify()}


# noinspection PyUnresolvedReferences
class ScriptPubKey(BaseScript, metaclass=ABCMeta):
    """
    Subclasses must either redefine `template` or reimplement their own `verify()` static method.
    They also must implement the `type` property.
    """
    template = None

    @classmethod
    def unhexlify(cls, hex_string, strict=True):
        if cls is ScriptPubKey:
            return cls(bytearray(unhexlify(hex_string)), strict=strict)
        else:
            return cls(Script(bytearray(unhexlify(hex_string))), strict=strict)

    @classmethod
    def verify(cls, bytes_):
        parser = ScriptParser(bytes_)
        if not bytes_:
            raise WrongScriptTypeException('Empty script')
        try:
            args = [data for data in parser.match(cls.template)]
        except UnexpectedOperationFound as exc:
            # print(exc)
            raise WrongScriptTypeException(str(exc))
        if len(args) == 1:
            return args[0]
        return args

    @staticmethod
    def empty():
        return ScriptPubKey(bytearray())

    def to_json(self, network=BitcoinMainnet):
        result = {'asm': str(self),
                  'hex': self.hexlify(),
                  'type': self.type}
        if self.address(network) is not None:
            result['address'] = str(self.address())
        return result

    @property
    def type(self):
        return 'scriptPubKey'

    @cached
    def p2sh_hash(self):
        ripemd160 = hashlib.new('ripemd160')
        ripemd160.update(hashlib.sha256(self.body).digest())
        return bytearray(ripemd160.digest())

    @cached
    def p2wsh_hash(self):
        return bytearray(sha256(self.body).digest())

    @cached
    def to_stack_data(self):
        return StackData.from_bytes(self.serialize())

    def is_standard(self):
        """Subclasses which have standard types should reimplement this method"""
        return False

    def address(self, network=BitcoinMainnet):
        """Subclasses which have a meaningful concept of address should reimplement this. For the moment
        we consider to have a meaningful address only for the following types: P2pkh, P2sh, P2wpkh, P2wsh"""
        return None


# noinspection PyUnresolvedReferences
class P2pkhScript(ScriptPubKey):

    template = 'OP_DUP OP_HASH160 <20> OP_EQUALVERIFY OP_CHECKSIG'

    compile_fmt = 'OP_DUP OP_HASH160 {} OP_EQUALVERIFY OP_CHECKSIG'

    def __init__(self, param, strict=True):
        """
        :param param: can be either of type `Script`, `PublicKey`, `Address` or `bytearray`.
        In the first case the script is verified and the public key hash is extracted.
        In the second case the script is built hashing the public key.
        In the third case, the script is built extracting the pubkeyhash from the address
        In the fourth case, the script is built from the public key hash (represented by the bytearray)
        """
        from .crypto import PublicKey
        if isinstance(param, Script):
            object.__setattr__(self, 'pubkeyhash', self.verify(param.body).data)
            super().__init__(param.body)
            # note: if we do not return here, we are going to call super().__init__ twice
            return

        if isinstance(param, PublicKey):
            object.__setattr__(self, 'pubkeyhash', param.hash())
        elif isinstance(param, bytearray):
            object.__setattr__(self, 'pubkeyhash', param)
        elif isinstance(param, P2pkhAddress):
            object.__setattr__(self, 'pubkeyhash', param.hash)
        else:
            raise TypeError('Wrong type for P2pkhScript __init__: {}'.format(type(param)))

        super().__init__(
            self.compile(
                self.compile_fmt.format(
                    hexlify(self.pubkeyhash).decode(),
                ),
            ),
            strict=strict,
        )

    def __repr__(self):
        return "P2pkhScript('{}')".format(hexlify(self.pubkeyhash).decode())

    @property
    def type(self):
        return 'p2pkh'

    def address(self, network=BitcoinMainnet):
        return P2pkhAddress.from_script(network, self)

    def is_standard(self):
        return True


class SegWitScript(ScriptPubKey, metaclass=ABCMeta):

    @staticmethod
    @abstractmethod
    def get_version():
        raise NotImplemented


class P2wpkhScript(P2pkhScript, SegWitScript, metaclass=ABCMeta):

    @staticmethod
    def get(segwit_version):
        for cls in P2wpkhScript.__subclasses__():
            if cls.get_version() == segwit_version:
                return cls
        raise ValueError('Undefined version: {}'.format(segwit_version))

    def address(self, network=BitcoinMainnet):
        return P2wpkhAddress.from_script(network, self)


class P2wpkhV0Script(P2wpkhScript):

    template = 'OP_0 <20>'

    compile_fmt = 'OP_0 {}'

    @staticmethod
    def get_version():
        return 0

    def __init__(self, param, strict=True):
        if isinstance(param, P2wpkhAddress):
            param = param.hash

        super().__init__(param, strict)

    def __repr__(self):
        return "P2wpkhScript('{}')".format(hexlify(self.pubkeyhash).decode())

    @property
    def type(self):
        return 'p2wpkhv0'

    def get_scriptcode(self):
        return P2pkhScript(self.pubkeyhash).to_stack_data()


# noinspection PyUnresolvedReferences
class P2shScript(ScriptPubKey):

    template = 'OP_HASH160 <20> OP_EQUAL'

    compile_fmt = 'OP_HASH160 {} OP_EQUAL'

    def __init__(self, param, strict=True):
        """
        :param param: can be either of type `Script`, `ScriptPubKey`, `Address` or `bytearray`.
        In the first case, the script is verified and the hash of the redeem script is extracted.
        In the second case, the script is built by hashing the redeem script.
        In the third case, the script is built by base58 decoding the address
        In the fourth case, the script is built from the script hash (represented by the bytearray)
        """
        if isinstance(param, Script):
            object.__setattr__(self, 'scripthash', self.verify(param.body).data)
            super().__init__(param.body, strict=strict)
            # note: if we do not return here, we are going to call super().__init__ twice
            return

        if isinstance(param, ScriptPubKey):
            object.__setattr__(self, 'scripthash', param.p2sh_hash())
        elif isinstance(param, P2shAddress):
            object.__setattr__(self, 'scripthash', param.hash)
        elif isinstance(param, bytearray):
            object.__setattr__(self, 'scripthash', param)
        else:
            raise TypeError('Wrong type for P2shScript __init__: {}'.format(type(param)))

        super().__init__(
            self.compile(
                self.compile_fmt.format(
                    hexlify(self.scripthash).decode(),
                ),
            ),
            strict=strict,
        )

    def __repr__(self):
        return "P2shScript('{}')".format(hexlify(self.scripthash).decode())

    @property
    def type(self):
        return 'p2sh'

    def is_standard(self):
        return True

    def address(self, network=BitcoinMainnet):
        return P2shAddress.from_script(network, self)


class P2wshScript(P2shScript, SegWitScript, metaclass=ABCMeta):

    @staticmethod
    def get(segwit_version):
        for cls in P2wshScript.__subclasses__():
            if cls.get_version() == segwit_version:
                return cls
        raise ValueError('Undefined version: {}'.format(segwit_version))

    def address(self, network=BitcoinMainnet):
        return P2wshAddress.from_script(network, self)


# noinspection PyUnresolvedReferences
class P2wshV0Script(P2wshScript):

    template = 'OP_0 <32>'

    compile_fmt = 'OP_0 {}'

    @staticmethod
    def get_version():
        return 0

    def __init__(self, param, strict=True):
        if isinstance(param, P2wpkhScript):
            raise ValueError("Can't embed P2WPKH script in P2WSH format")
        # segwit p2wsh have different hashing method than regular p2sh scripts!
        if isinstance(param, ScriptPubKey):
            param = param.p2wsh_hash()
        if isinstance(param, P2wshAddress):
            param = param.hash

        super().__init__(param, strict)

    def __repr__(self):
        return "P2wshV0Script('{}')".format(hexlify(self.scripthash).decode())

    @property
    def type(self):
        return 'p2wshv0'


# noinspection PyUnresolvedReferences
class P2pkScript(ScriptPubKey):

    template = '<33|65> OP_CHECKSIG'

    def __init__(self, param, strict=True):
        """
        :param param: can be either of type `Script` or `PublicKey`.
        In the first case it is verified and the public key is extracted.
        In the second case the script is built from the public key
        """
        if isinstance(param, Script):
            pubkey = self.verify(param.body)

            try:
                object.__setattr__(self, 'pubkey', PublicKey(pubkey.data))
            except WrongPubKeyFormat:
                if strict:
                    raise
                object.__setattr__(self, 'pubkey', pubkey)

            super().__init__(param.body, strict=strict)
        elif isinstance(param, PublicKey):
            object.__setattr__(self, 'pubkey', param)
            super().__init__(
                self.compile('{} OP_CHECKSIG'.format(self.pubkey.hexlify())),
                strict=strict,
            )
        elif isinstance(param, StackData):
            if strict:
                raise TypeError('Must provide an object of type PublicKey in strict mode')
            if len(param) not in (33, 65):
                raise WrongScriptTypeException('Public keys must be either 33 or 65 bytes long')
            object.__setattr__(self, 'pubkey', param)
        else:
            raise TypeError('Wrong type for P2pkScript __init__: {}'.format(type(param)))

    def __repr__(self):
        return "P2pkScript('{}')".format(self.pubkey)

    @property
    def type(self):
        return 'p2pk'

    def is_standard(self):
        return True


# noinspection PyUnresolvedReferences
class NulldataScript(ScriptPubKey):

    template = 'OP_RETURN <1-83>'
    max_nulldata_data_size = 80

    def __init__(self, param, strict=True):
        """
        :param param: can be either of type `Script` or `StackData`.
        In the first case the script is verified and the data is extracted.
        In the second case the script is built from the data provided.
        """
        if isinstance(param, Script):
            object.__setattr__(self, 'data', self.verify(param.body))
            super().__init__(param.body, strict=strict)
        elif isinstance(param, StackData):
            object.__setattr__(self, 'data', param)
            super().__init__(
                self.compile('OP_RETURN {}'.format(param)),
                strict=strict,
            )
        else:
            raise TypeError('Wrong type for NulldataScript __init__: {}'.format(type(param)))

    def __repr__(self):
        return "NulldataScript('{}')".format(self.data.hexlify())

    @property
    def type(self):
        return 'nulldata'

    def is_standard(self):
        # 1-byte OP_RETURN + 2-byte pushdata + 80-byte data
        return len(self.data) <= self.max_nulldata_data_size and len(self.body) <= self.max_nulldata_data_size + 3


# noinspection PyUnresolvedReferences
class MultisigScript(ScriptPubKey):

    template = '<>+ OP_CHECKMULTISIG'

    @staticmethod
    def _parse_pubkeys(pubkeys):
        result = []
        for pubkey in pubkeys:
            try:
                result.append(PublicKey(pubkey.data))
            except WrongPubKeyFormat:
                result.append(pubkey)
        return result

    @staticmethod
    def _verify_pubkeys(pubkeys):
        valid = 0
        for pubkey in pubkeys:
            if isinstance(pubkey, PublicKey):
                valid += 1
            elif isinstance(pubkey, StackData):
                if len(pubkey) not in (33, 65):
                    raise WrongScriptTypeException('Multisig public keys must be either 33 or 65 bytes long')
            else:
                raise TypeError('Passed type {} as public key'.format(type(pubkey)))
        return valid

    @classmethod
    def verify(cls, bytes_, strict=True):

        parser = ScriptParser(bytes_)
        if not bytes_:
            raise WrongScriptTypeException('Empty script')
        try:
            push_ops = parser.match(cls.template)
            if len(push_ops) <= 2:
                raise WrongScriptTypeException('Less than 3 push ops before OP_CHECKMULTISIG')
            m, *pubkeys, n = [data for data in push_ops]
        except UnexpectedOperationFound as exc:
            raise WrongScriptTypeException(str(exc))

        if len(pubkeys) == 0 or len(pubkeys) != int(n):
            raise WrongScriptTypeException('Non-matching N and number of pubkeys')

        pubkeys = cls._parse_pubkeys(pubkeys)
        valid = cls._verify_pubkeys(pubkeys)

        if strict and valid < int(m):
            raise WrongPubKeyFormat('{} valid public keys while m is {}'.format(valid, int(m)))

        return [int(m)] + pubkeys + [int(n)]

    def __init__(self, *args, strict=True):
        """
        :param args: if one arg is provided that is interpreted as a precompiled script which needs
        verification to see if it belongs to this type. Once verification is done, `m`, a list of pubkeys
        and `n` are extracted and saved.
        If more than one arg is provided, we assume that the parameters are `m, pubkey1, ..., pubkeyn, n`.
        """

        if len(args) == 0:
            raise TypeError('Wrong number of params for MultisigScript __init__: {}'.format(len(args)))
        if len(args) == 1:
            # we expect something of type Script
            script = args[0]
            super().__init__(script.body, strict=strict)
            m, *pubkeys, n = self.verify(script.body, strict=strict)
        else:
            m, *pubkeys, n = args

            valid = self._verify_pubkeys(pubkeys)
            if strict and valid < m:
                raise WrongPubKeyFormat('{} valid public keys while m is {}'.format(valid, m))

        if n != len(pubkeys):
            raise ValueError('Pushed {} keys but n is {}'.format(len(pubkeys), n))

        object.__setattr__(self, 'm', m)
        object.__setattr__(self, 'pubkeys', pubkeys)
        object.__setattr__(self, 'n', n)

        if len(args) != 1:
            # in this case we haven't called super().__init__() yet
            super().__init__(
                self.compile(
                    '{} {} {} OP_CHECKMULTISIG'.format(
                        StackData.from_int(self.m),
                        ' '.join([pubk.hexlify() for pubk in self.pubkeys]),
                        StackData.from_int(len(self.pubkeys)),
                    ),
                ),
                strict=strict,
            )

    def __repr__(self):
        return "MultisigScript('{}', {}, '{}')".format(self.m,
                                                       ', '.join("'{}'".format(pk) for pk in self.pubkeys),
                                                       self.n)

    @property
    def type(self):
        return 'multisig'

    @property
    def req_sigs(self):
        return self.m

    def is_standard(self):
        return (1 <= self.n <= 3) and (1 <= self.m <= self.n)


# noinspection PyUnresolvedReferences
class IfElseScript(ScriptPubKey):

    @staticmethod
    def verify(bytes_):
        try:
            parser = ScriptParser(bytes_)
            if_counter = 0
            if_script = Stream()
            else_script = Stream()
            current_script = if_script
            else_found = False

            try:
                if parser[-1] != OpCodeConverter.to_int('OP_ENDIF'):
                    raise WrongScriptTypeException('Script is not OP_ENDIF terminated')
            except IndexError:
                raise WrongScriptTypeException

            parser.require('OP_IF')
            # print('################### Parsing IfElseScript ######################')
            for op in parser:
                # print('Parsing op: {}'.format(op))
                if op == OpCodeConverter.to_int('OP_IF'):
                    current_script << op
                    if_counter += 1
                elif op == OpCodeConverter.to_int('OP_ENDIF'):
                    if_counter -= 1
                    if if_counter == -1:
                        break
                    current_script << op
                elif op == OpCodeConverter.to_int('OP_ELSE'):
                    if if_counter == 0:
                        # switch script
                        else_found = True
                        # print('Switching script. If script is: {}'.format(if_script.hexlify()))
                        current_script = else_script
                    else:
                        current_script << op
                else:
                    try:
                        push_data = parser.get_push(op)
                        current_script << push_data.to_push_op()
                    except UnexpectedOperationFound:
                        # not a push operation, we can move on
                        current_script << op

            if not else_found:
                raise WrongScriptTypeException('No OP_ELSE found matching outer OP_IF')

            parser.require_empty()
            # print('IF SCRIPT: {}'.format(if_script.hexlify()))
            # print('ELSE SCRIPT: {}'.format(else_script.hexlify()))
            return (ScriptBuilder.identify(if_script.serialize(), inner=True),
                    ScriptBuilder.identify(else_script.serialize(), inner=True))
        except UnexpectedOperationFound as exc:
            raise WrongScriptTypeException(str(exc))

    # noinspection PyMissingConstructor
    def __init__(self, *args, strict=True):
        """
        :param args: if one arg is provided, it is interpreted as a script, which is in turn
        verified and if-branch and else-branch are extracted. If two args are provided they
        are supposed to be the if-branch and else-branch. In this case the script is created
        from these data
        """

        if len(args) == 1:
            # we expect something of type Script
            script = args[0]
            if_script, else_script = self.verify(script.body)
            object.__setattr__(self, 'if_script', if_script)
            object.__setattr__(self, 'else_script', else_script)
            super().__init__(script.body, strict=strict)
        elif len(args) == 2:
            if_script, else_script = args
            object.__setattr__(self, 'if_script', if_script)
            object.__setattr__(self, 'else_script', else_script)
            super().__init__(
                bytearray([OpCodeConverter.to_int('OP_IF')]) +
                self.if_script.serialize() +
                bytearray([OpCodeConverter.to_int('OP_ELSE')]) +
                self.else_script.serialize() +
                bytearray([OpCodeConverter.to_int('OP_ENDIF')]),
                strict=strict,
            )
        else:
            raise TypeError('Wrong number of params for IfElseScript __init__: {}'.format(len(args)))

    def __repr__(self):
        return 'IfElseScript({}, {})'.format(self.if_script, self.else_script)

    @property
    def type(self):
        return 'if{{ {} }}else{{ {} }}'.format(self.if_script.type, self.else_script.type)


# noinspection PyUnresolvedReferences
class TimelockScript(ScriptPubKey):

    @staticmethod
    def verify(bytes_):
        try:
            parser = ScriptParser(bytes_)
            locktime = int(parser.get_push())
            parser.require('OP_CHECKLOCKTIMEVERIFY')
            parser.require('OP_DROP')
            script = parser >> len(parser)
            return locktime, ScriptBuilder.identify(script, inner=True)
        except (UnexpectedOperationFound, StopIteration, IndexError) as exc:
            raise WrongScriptTypeException(str(exc))

    def __init__(self, *args, strict=True):
        """
        :param args: if one arg is provided it is interpreted as a script, which is in turn
        verified and `locktime` and `locked_script` are extracted. If two args are provided,
        they are interpreted as `locktime` and `locked_script` respectively, the script is
        then generated from these params
        """
        from .transaction import Locktime
        if len(args) == 1:
            script = args[0]
            locktime, locked_script = self.verify(script.body)
            object.__setattr__(self, 'locked_script', locked_script)
            object.__setattr__(self, 'locktime', locktime)

            super().__init__(script.body, strict=strict)
        elif len(args) == 2:
            locktime, locked_script = args
            if not isinstance(locktime, Locktime):
                raise TypeError('locktime is not of type Locktime, {} instead'.format(type(locktime)))
            if not isinstance(locked_script, BaseScript):
                raise TypeError('locked_script is not of type Script, {} instead'.format(type(locked_script)))
            object.__setattr__(self, 'locked_script', locked_script)
            object.__setattr__(self, 'locktime', locktime)

            script_body = Stream()
            script_body << self.locktime.for_script().to_push_op()
            script_body << OpCodeConverter.to_int('OP_CHECKLOCKTIMEVERIFY')
            script_body << OpCodeConverter.to_int('OP_DROP')
            script_body << self.locked_script

            super().__init__(script_body.serialize(), strict=strict)

        else:
            raise TypeError('Wrong number of params for TimelockScript __init__: {}'.format(len(args)))

    def __repr__(self):
        return 'TimelockScript({}, {})'.format(self.locktime, self.locked_script)

    @property
    def type(self):
        return '[timelock] {}'.format(self.locked_script.type)


# noinspection PyUnresolvedReferences
class RelativeTimelockScript(ScriptPubKey):

    @staticmethod
    def verify(bytes_):
        try:
            parser = ScriptParser(bytes_)
            sequence = parser.get_push()
            parser.require('OP_CHECKSEQUENCEVERIFY')
            parser.require('OP_DROP')
            script = parser >> len(parser)
            return sequence, ScriptBuilder.identify(script, inner=True)
        except (UnexpectedOperationFound, StopIteration) as exc:
            raise WrongScriptTypeException(str(exc))

    def __init__(self, *args, strict):
        """
        :param args: if one arg is provided, it is interpreted as a script, which is in turn
        verified and `sequence` and `locked_script` are extracted. If two args are provided,
        they are interpreted as `sequence` and `locked_script` respectively, the script is
        then generated from these params
        """
        from .transaction import Sequence
        if len(args) == 1:
            script = args[0]
            sequence, locked_script = self.verify(script.body)
            object.__setattr__(self, 'locked_script', locked_script)
            object.__setattr__(self, 'sequence', Sequence(int(sequence)))
            super().__init__(script.body, strict=strict)
        elif len(args) == 2:
            sequence, locked_script = args
            object.__setattr__(self, 'sequence', sequence)
            object.__setattr__(self, 'locked_script', locked_script)
            script_body = Stream()
            script_body << self.sequence.for_script().to_push_op()
            script_body << OpCodeConverter.to_int('OP_CHECKSEQUENCEVERIFY')
            script_body << OpCodeConverter.to_int('OP_DROP')
            script_body << self.locked_script

            super().__init__(script_body.serialize(), strict=strict)
        else:
            raise TypeError('Wrong number of params for RelativeTimelockScript __init__: {}'.format(len(args)))

    def __repr__(self):
        return 'RelativeTimelockScript({}, {})'.format(self.sequence, self.locked_script)

    @property
    def type(self):
        return '[relativetimelock] {}'.format(self.locked_script.type)


class HashlockScript(ScriptPubKey):

    @staticmethod
    @abstractmethod
    def hash_func(data):
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def hash_op():
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def hash_size():
        raise NotImplemented

    @classmethod
    def from_preimage(cls, preimage, script):
        return cls([StackData.from_bytes(cls.hash_func(preimage)), script])

    @classmethod
    def verify(cls, bytes_):
        try:
            parser = ScriptParser(bytes_)
            parser.require(cls.hash_op())
            secret_hash = parser.require_push(str(cls.hash_size()))
            parser.require('OP_EQUALVERIFY')
            script = parser >> len(parser)
            return secret_hash, ScriptBuilder.identify(script, inner=True)
        except (UnexpectedOperationFound, StopIteration) as exc:
            raise WrongScriptTypeException(str(exc))

    def __init__(self, *args, strict=True):
        """
        :param args: if one arg is provided, it is interpreted as a script, which is in turn
        verified and `hash` and `locked_script` are extracted. If two args are provided,
        they are interpreted as `hash` and `locked_script` respectively, the script is
        then generated from these params
        """

        if len(args) == 1:
            script = args[0]
            lock_hash, locked_script = self.verify(script.body)
            object.__setattr__(self, 'locked_script', locked_script)
            if not isinstance(lock_hash, StackData):
                lock_hash = StackData.from_bytes(lock_hash)
            object.__setattr__(self, 'hash', lock_hash)
            super().__init__(script.body, strict=strict)
        elif len(args) == 2:
            lock_hash, locked_script = args
            object.__setattr__(self, 'locked_script', locked_script)
            if not isinstance(lock_hash, StackData):
                lock_hash = StackData.from_bytes(lock_hash)
            object.__setattr__(self, 'hash', lock_hash)
            script_body = Stream()
            script_body << OpCodeConverter.to_int(self.__class__.hash_op())
            script_body << self.hash.to_push_op()
            script_body << OpCodeConverter.to_int('OP_EQUALVERIFY')
            script_body << self.locked_script
            super().__init__(script_body.serialize(), strict=strict)
        else:
            raise TypeError('Wrong number of params for HashlockScript __init__: {}'.format(len(args)))

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.hash.hexlify(), self.locked_script)

    @property
    def type(self):
        return '[hashlock] {}'.format(self.locked_script.type)


class Hashlock256Script(HashlockScript):

    @staticmethod
    def hash_op():
        return 'OP_HASH256'

    @staticmethod
    def hash_func(data):
        return bytearray(sha256(sha256(data).digest()).digest())

    @staticmethod
    def hash_size():
        return 32


class Hashlock160Script(HashlockScript):

    @staticmethod
    def hash_op():
        return 'OP_HASH160'

    @staticmethod
    def hash_func(data):
        sha = sha256(data).digest()
        ripe = hashlib.new('ripemd160')
        ripe.update(sha)
        return bytearray(ripe.digest())

    @staticmethod
    def hash_size():
        return 20


class UnknownScript(ScriptPubKey):

    @property
    def type(self):
        return 'nonstandard'


class ScriptBuilder(object):

    # never change the order of the elements in this list
    types = [P2pkhScript,
             P2shScript,
             NulldataScript,
             P2pkScript,
             MultisigScript,
             IfElseScript,
             RelativeTimelockScript,
             TimelockScript,
             Hashlock256Script,
             Hashlock160Script,
             P2wpkhV0Script,
             P2wshV0Script]

    not_allowed_inner = {P2wshV0Script, P2wpkhV0Script, P2shScript}  # can't be an internal ScriptPubKey format
    not_allowed_redeem = {P2shScript}  # can't be an external RedeemScript format

    @staticmethod
    def identify(raw_script, inner=False, redeem=False, strict=True):

        if isinstance(raw_script, str):
            raw_script = bytearray(unhexlify(raw_script))

        if inner:
            ignore = ScriptBuilder.not_allowed_inner
        elif redeem:
            ignore = ScriptBuilder.not_allowed_redeem
        else:
            ignore = set()

        for script_type in [script for script in ScriptBuilder.types if script not in ignore]:
            try:    
                # print('Trying {}...'.format(script_type.__name__))
                candidate = script_type(Script(raw_script, strict=strict), strict=strict)
                # print('Success')
                return candidate
            except (WrongScriptTypeException, WrongPubKeyFormat, WrongPushDataOp) as exc:
                pass
        return UnknownScript(raw_script)

"""
# Example usage:

from .crypto import PublicKey
from .transaction import Sequence

P2wshV0Script(
    IfElseScript(
        MultisigScript(
            2,
            PublicKey.unhexlify("021b98b2e4ba9dae9f869bcf948c45df6b6f8e6bb623915cf144237f5e6ab98cf4"),
            PublicKey.unhexlify("0376d53363bbeefed905fc685e4d4e1fe0cbf9959e8f59e9f5f209f489b3a62857"),
            2
        ),
        RelativeTimelockScript(
            Sequence(5),
            P2pkScript(
                PublicKey.unhexlify("0376d53363bbeefed905fc685e4d4e1fe0cbf9959e8f59e9f5f209f489b3a62857")
            )
        )
    )
)
"""



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.


import unittest
from unittest.mock import patch

from btcpy.constants import (
    BitcoinMainnet,
    BitcoinTestnet,
    PeercoinMainnet,
    PeercoinTestnet,
)
from btcpy.structs.transaction import *
from btcpy.structs.script import *
from btcpy.structs.block import *
from btcpy.structs.crypto import PublicKey, PrivateKey
from btcpy.structs.address import Address, SegWitAddress, P2shAddress, P2wshAddress
from btcpy.lib.codecs import CouldNotDecode
from btcpy.structs.hd import *
from btcpy.lib.base58 import (
    b58decode,
    b58decode_check,
    b58encode,
    b58encode_check,
)
from btcpy.lib.parsing import IncompleteParsingException


def get_data(filename):
    import os
    import json
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open('{}/data/{}.json'.format(dir_path, filename)) as infile:
        return json.load(infile)


transactions = get_data('rawtxs')
scripts = get_data('scripts')
unknownscripts = get_data('unknownscripts')
keys = get_data('xkeys')
segwit_valid_addresses = get_data('segwit_addr_valid')
segwit_invalid_addresses = get_data('segwit_addr_invalid')
valid_blocks = get_data('valid_blocks')
invalid_blocks = get_data('invalid_blocks')
short_blocks = get_data('short_blocks')
hd_keys = get_data('hd')
addresses = get_data('addr')
malleated_txs = get_data('malleated')
privk = get_data('priv_addr_path')
segsig = get_data('segwitsig')
p2sh = get_data('p2sh')
priv_pub_hash_addr_p2pkh_segwit = get_data('priv_pub_hash_addr_p2pkh_segwit')
b58 = get_data('base58')
b58chk = get_data('base58_check')
segwit_hashes = get_data('segwit_hashes')
wif = get_data('wif')
p2wpkh_over_p2sh = get_data("p2wpkh_over_p2sh")
p2wsh_over_p2sh = get_data("p2wsh_over_p2sh")
serialization_data = get_data('stack_data/serialization')


class TestB58(unittest.TestCase):

    def test_b58encode(self):
        for hexa, encoded in b58:
            self.assertEqual(b58encode(unhexlify(hexa)), encoded)

    def test_b58encode_check(self):
        for hexa, encoded in b58chk:
            self.assertEqual(b58encode_check(unhexlify(hexa)), encoded)

    def test_b58decode(self):
        for hexa, encoded in b58:
            self.assertEqual(hexlify(b58decode(encoded)).decode(), hexa)

    def test_b58decode_check(self):
        for hexa, encoded in b58chk:
            self.assertEqual(hexlify(b58decode_check(encoded)).decode(), hexa)


class TestSegwitHashes(unittest.TestCase):

    def test_hashes(self):
        for tx in segwit_hashes:
            parsed = Transaction.unhexlify(tx['tx'])
            self.assertEqual(parsed.txid, tx['txid'])
            self.assertEqual(parsed.hash(), tx['hash'])


class TestUnknownScript(unittest.TestCase):

    def test(self):
        for script in unknownscripts:
            result = ScriptBuilder.identify(unhexlify(script['hex']))
            self.assertTrue(isinstance(result, UnknownScript))
            self.assertEqual(result.hexlify(), script['hex'])
            self.assertEqual(str(result), script['asm'])


class TestPrivPubHashAddrP2pkhSegwit(unittest.TestCase):

    def test(self):
        for data in priv_pub_hash_addr_p2pkh_segwit:
            try:
                net = BitcoinMainnet
                priv = PrivateKey.from_wif(net, data['privkey'])
            except ValueError:
                net = BitcoinTestnet
                priv = PrivateKey.from_wif(net, data['privkey'])

            pub = PublicKey.unhexlify(data['pubkey'])
            pubhash = bytearray(unhexlify(data['pubkeyhash']))
            address = Address.from_string(net, data['address'])
            p2pkhhex = data['scriptpubkey']
            segwit_addr = data['segwit']

            self.assertEqual(priv.pub(), pub)
            self.assertEqual(pub.hash(), pubhash)
            self.assertEqual(address.hash, pubhash)
            self.assertEqual(P2pkhScript(pub).hexlify(), p2pkhhex)
            self.assertEqual(P2pkhScript(address).hexlify(), p2pkhhex)
            self.assertEqual(P2pkhScript(pubhash).hexlify(), p2pkhhex)
            self.assertEqual(str(P2shScript(P2wpkhV0Script(pub)).address(net)), segwit_addr)
            self.assertEqual(str(P2shScript(P2wpkhV0Script(pubhash)).address(net)), segwit_addr)
            self.assertEqual(P2shScript(P2wpkhV0Script(pub)).scripthash, Address.from_string(net, segwit_addr).hash)
            self.assertEqual(P2shScript(P2wpkhV0Script(pubhash)).scripthash, Address.from_string(net, segwit_addr).hash)


class TestNormalizedId(unittest.TestCase):

    def test(self):
        for tx1, tx2 in malleated_txs:
            tx1 = Transaction.unhexlify(tx1)
            tx2 = Transaction.unhexlify(tx2)
            self.assertNotEqual(tx1.txid, tx2.txid)
            self.assertEqual(tx1.normalized_id, tx2.normalized_id)


class TestHD(unittest.TestCase):

    def test_hd(self):
        masterpriv = None
        for data in hd_keys:
            priv = ExtendedKey.decode(BitcoinMainnet, data['prv'])
            pub = ExtendedKey.decode(BitcoinMainnet, data['pub'])
            if data['path'] == 'm':
                masterpriv = priv
            self.assertEqual(priv.pub().encode(BitcoinMainnet), pub.encode(BitcoinMainnet))
            self.assertEqual(priv.encode(BitcoinMainnet), data['prv'])
            self.assertEqual(pub.encode(BitcoinMainnet), data['pub'])
            derived = masterpriv.derive(data['path'])
            self.assertEqual(derived.encode(BitcoinMainnet), data['prv'])
            self.assertEqual(derived.pub().encode(BitcoinMainnet), data['pub'])

    def test_priv_pub(self):
        masterpub = ExtendedPublicKey.decode(BitcoinMainnet, hd_keys[0]['pub'])
        masterpriv = ExtendedPublicKey.decode(BitcoinMainnet, hd_keys[0]['prv'])
        pubs = [masterpub]
        privs = [masterpriv]
        paths = ['m/0/1/2147483646/2',
                 './2147483646/0',
                 './0/2147483647/1/2147483646/2',
                 './156131385/44645489/4865448/4896853']
        for path in paths:
            newpubs = []
            newprivs = []
            for pub, priv in zip(pubs, privs):
                newpubs.append(pub.derive(path))
                newprivs.append(priv.derive(path))
                self.assertEqual(newpubs[-1], newprivs[-1].pub())
            pubs += newpubs
            privs += newprivs


class TestBlock(unittest.TestCase):
    def test_block_deserialize_serialize(self):
        for i in range(len(valid_blocks)):
            unhex_block = Block.unhexlify(valid_blocks[i]['raw'])
            serial_block = unhex_block.serialize()
            self.assertEqual(valid_blocks[i]['raw'], hexlify(serial_block).decode())

    def test_block_hash(self):
        for i in range(len(valid_blocks)):
            unhex_block = Block.unhexlify(valid_blocks[i]['raw'])
            self.assertEqual(valid_blocks[i]['hash'], unhex_block.hash())

    def test_block_header(self):
        for i in range(len(valid_blocks)):
            unhex_block = Block.unhexlify(valid_blocks[i]['raw'])
            unhex_header = BlockHeader.unhexlify(valid_blocks[i]['raw'])
            self.assertEqual(hexlify(unhex_block.header.serialize()), hexlify(unhex_header.serialize()))

    def test_fail_block_deserialize_serialize(self):
        for i in range(len(valid_blocks)):
            unhex_block = Block.unhexlify(invalid_blocks[i]['raw'])
            serial_block = unhex_block.serialize()
            self.assertNotEqual(valid_blocks[i]['raw'], hexlify(serial_block).decode())

    def test_fail_block_hash(self):
        for i in range(len(valid_blocks)):
            unhex_block = Block.unhexlify(invalid_blocks[i]['raw'])
            self.assertNotEqual(valid_blocks[i]['hash'], unhex_block.hash())

    def test_fail_block_header(self):
        for i in range(len(valid_blocks)):
            unhex_block = Block.unhexlify(valid_blocks[i]['raw'])
            unhex_header = BlockHeader.unhexlify(invalid_blocks[i]['raw'])
            self.assertNotEqual(hexlify(unhex_block.header.serialize()), hexlify(unhex_header.serialize()))

    def test_stop_iteration(self):
        for i in range(len(short_blocks)):
            with self.assertRaises(StopIteration):
                Block.unhexlify(short_blocks[i]['raw'])
            with self.assertRaises(StopIteration):
                Block.unhexlify(short_blocks[i]['raw']).serialize()

    def test_empty_deserialized_string(self):
        for i in range(len(valid_blocks)):
            parser = BlockParser(bytearray(unhexlify(valid_blocks[i]['raw'])), TransactionParser, BitcoinMainnet)
            parser.get_block_header()
            parser.get_txns()
            with self.assertRaises(StopIteration):
                parser >> 1

    def test_incomplete_parsing_exception(self):
        for i in range(len(valid_blocks)):
            aug_raw = valid_blocks[i]['raw'] + "ff"
            with self.assertRaises(IncompleteParsingException):
                Block.unhexlify(aug_raw)


class TestTransaction(unittest.TestCase):

    def test_serialization(self):
        for data in transactions:
            tx = Transaction.unhexlify(data['raw'])
            computed = tx.hexlify()
            original = data['raw']
            self.assertEqual(computed, original)

    def test_txid(self):
        for data in transactions:
            self.assertEqual(Transaction.unhexlify(data['raw']).txid, data['txid'])

    def test_script(self):
        for key, value in scripts.items():
            self.assertEqual(eval(value['code']).hexlify(), value['hex'])
            self.assertEqual(str(eval(value['code'])), value['asm'])
            parsed_script = Script.unhexlify(value['hex'])
            self.assertEqual(parsed_script.hexlify(), value['hex'])
            self.assertEqual(str(parsed_script), value['asm'])
            parsed_script = ScriptBuilder.identify(value['hex'])
            self.assertEqual(parsed_script.hexlify(), value['hex'])
            self.assertEqual(str(parsed_script), value['asm'])
            self.assertEqual(parsed_script.type, value['type'])

    # Fails on the last transaction for some reason :\
    def test_json(self):
        for data in transactions:
            tx = Transaction.unhexlify(data["raw"])
            tx_json = tx.to_json()
            tx_from_json = Transaction.from_json(tx_json)
            self.assertEqual(tx, tx_from_json)


class TestSegWitAddress(unittest.TestCase):

    def test_valid(self):
        for data in segwit_valid_addresses:
            try:
                net = BitcoinMainnet
                address = SegWitAddress.from_string(net, data['address'])
            except CouldNotDecode:
                net = BitcoinTestnet
                address = SegWitAddress.from_string(net, data['address'])

            script = ScriptBuilder.identify(data['script'])
            self.assertEqual(address.hash, script.address(net).hash)
            if len(data['script']) == 44:
                self.assertEqual(P2wpkhV0Script(address), script)
            elif len(data['script']) == 68:
                self.assertEqual(P2wshV0Script(address), script)

    def test_invalid(self):
        for address in segwit_invalid_addresses:
            with self.assertRaises(CouldNotDecode):
                print(SegWitAddress.from_string(BitcoinMainnet, address))


class TestSegwitOverP2sh(unittest.TestCase):

    def test_p2wpkh_over_p2sh(self):
        for spend in p2wpkh_over_p2sh:
            pubkey = PublicKey.unhexlify(spend['pubkey'])
            self.assertEqual(str(P2shAddress.from_script(BitcoinMainnet, P2wpkhScript.get(spend['witness_version'])(pubkey),)), spend['address'])
            self.assertEqual(P2wpkhScript.get(spend['witness_version'])(pubkey).hexlify(), spend['redeem_script'])
            self.assertEqual(str(P2shScript(P2wpkhScript.get(spend['witness_version'])(pubkey))), spend['script_pubkey'])

    def test_p2wsh_over_p2sh(self):
        for spend in p2wsh_over_p2sh:
            wit_script = ScriptBuilder.identify(spend['witness_script'])
            self.assertEqual(str(P2shScript(P2wshScript.get(spend['witness_version'])(wit_script))), spend['script_pubkey'])
            self.assertEqual(P2wshScript.get(spend['witness_version'])(wit_script).hexlify(), spend['redeem_script'])
            self.assertEqual(str(P2shAddress.from_script(BitcoinMainnet, P2wshScript.get(spend['witness_version'])(wit_script))), spend['address'])


class TestReplace(unittest.TestCase):

    def test_success(self):
        for transaction in transactions:
            tx = MutableTransaction.unhexlify(transaction['raw'])
            self.assertEqual(tx.is_replaceable(), transaction['replaceable'])
            if len(tx.ins) > 2:
                tx.ins[1].sequence = Sequence(0xfffffffd)
                self.assertTrue(tx.ins[1].is_replaceable())
                self.assertFalse(tx.ins[2].is_replaceable())
                self.assertTrue(tx.is_replaceable())


class TestStackData(unittest.TestCase):

    @staticmethod
    def get_test_data():
        from os import walk
        files = next(walk('./stack_data/data/rand_data'))[2]
        for file in files:
            with open('stack_data/data/rand_data/'+file, 'rb') as infile:
                yield bytearray(infile.read())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fail_sizes = [2**32 + 1]

    def test_success(self):
        stored_data = TestStackData.get_test_data()
        for data in stored_data:
            stack = StackData.from_bytes(data)
            push_op = stack.to_push_op()
            if len(data) == 1 and 0 <= data[0] <= 16:
                self.assertTrue(len(push_op) == len(data))
            elif 1 <= len(data) <= 75:
                self.assertTrue(len(push_op) == len(data) + 1)
                self.assertTrue(push_op[0] == len(data))
            elif 76 <= len(data) <= 255:
                self.assertTrue(len(push_op) == len(data) + 2)
                self.assertTrue(push_op[0] == 76)
                self.assertTrue(push_op[1] == len(data))
            elif 256 <= len(data) <= 2**16 - 1:
                self.assertTrue(len(push_op) == len(data) + 3)
                self.assertTrue(push_op[0] == 77)
                self.assertTrue(int.from_bytes(push_op[1:3], 'little') == len(data))
            elif 65536 <= len(data) <= 2**32 - 1:
                self.assertTrue(len(push_op) == len(data) + 5)
                self.assertTrue(push_op[0] == 78)
                self.assertTrue(int.from_bytes(push_op[1:5], 'little') == len(data))

    def test_basic(self):
        for item in serialization_data:
            if 'int' in item:
                data = StackData.from_int(item['int'])
            else:
                push_op, data = item['data']
                data = StackData(push_op, unhexlify(data))
            self.assertEqual(Witness([data]).hexlify(), item['hex'])

    def test_failure(self):
        for size in self.fail_sizes:
            with patch('btcpy.structs.script.len', return_value=size, create=True):
                with self.assertRaises(WrongPushDataOp):
                    StackData(bytearray([0]))

    def test_int_conversion(self):
        import os
        path = os.path.realpath(__file__).rsplit('/', 1)[0]
        with open('{}/data/stack_data/rand_ints'.format(path)) as infile:
            ints = [int(x) for x in infile.read().split()]
        ints = list(range(1001)) + ints
        for x in ints:
            self.assertTrue(int(StackData.from_int(x)) == x)


class TestKeys(unittest.TestCase):

    def test_priv_to_pubhash(self):
        for priv, addr, _ in privk['derivations']:     
            self.assertEqual(str(PrivateKey.from_wif(BitcoinTestnet, priv).pub().to_address(BitcoinTestnet)), addr)
            self.assertEqual(PrivateKey.from_wif(BitcoinTestnet, priv).pub().hash(),
                             Address.from_string(BitcoinTestnet, addr).hash)

    def test_derivation(self):
        m = ExtendedPrivateKey.decode(BitcoinTestnet, privk['master'])
        for priv, _, path in privk['derivations']:
            self.assertEqual(m.derive(path).key, PrivateKey.from_wif(BitcoinTestnet, priv))

    def test_to_wif(self):
        for w in wif:
            try:
                net = BitcoinMainnet
                self.assertEqual(PrivateKey.from_wif(net, w['wif']).hexlify(), w['hex'])
            except:
                net = BitcoinTestnet
                self.assertEqual(PrivateKey.from_wif(net, w['wif']).hexlify(), w['hex'])

            priv = PrivateKey.unhexlify(w['hex'])
            if not w['compressed']:
                priv.public_compressed = False
            self.assertEqual(priv.to_wif(net), w['wif'])


class TestPubkey(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.compressed = '03fe43d0c2c3daab30f9472beb5b767be020b81c7cc940ed7a7e910f0c1d9feef1'
        self.uncompressed = ('04fe43d0c2c3daab30f9472beb5b767be020b81c7cc940ed7a7e910f0c1d9feef10fe85eb3ce193405c2dd845'
                             '3b7aeb6c1752361efdbf4f52ea8bf8f304aab37ab')
        self.address = '1KKKK6N21XKo48zWKuQKXdvSsCf95ibHFa'

    def test_compression(self):
        pubk = PublicKey.unhexlify(self.uncompressed)
        self.assertTrue(hexlify(pubk.compressed).decode() == self.compressed)

    def test_uncompression(self):
        pubk = PublicKey.unhexlify(self.compressed)
        self.assertTrue(hexlify(pubk.uncompressed).decode() == self.uncompressed)

    def test_address_generation(self):
        pubk = PublicKey.unhexlify(self.uncompressed)
        self.assertTrue(str(pubk.to_address(BitcoinMainnet)) == self.address)


class TestPrivKey(unittest.TestCase):

    def setUp(self):
        from ecdsa import SigningKey, SECP256k1

        self.mainnet_privs = []
        self.testnet_privs = []
        self.mainnet_vers = []
        self.testne_vers = []

        for k in keys:
            try:
                self.mainnet_privs.append(ExtendedPrivateKey.decode(BitcoinMainnet, k[1]).key)
            except ValueError:
                self.testnet_privs.append(ExtendedPrivateKey.decode(BitcoinTestnet, k[1]).key)

        self.mainnet_vers = [SigningKey.from_string(p.key, curve=SECP256k1).get_verifying_key() for p in self.mainnet_privs]

        self.testnet_vers = [SigningKey.from_string(p.key, curve=SECP256k1).get_verifying_key() for p in self.testnet_privs]

    def test_raw_sig_success(self):
        from os import urandom
        for _ in range(10):

            for ver, priv in zip(self.mainnet_vers, self.mainnet_privs):
                digest = bytearray(urandom(32))
                r, s, order = priv.raw_sign(digest)
                self.assertTrue(s in range(1, PrivateKey.highest_s))
                self.assertTrue(ver.verify_digest((r, s), digest, sigdecode=lambda sig, _: (sig[0], sig[1])))

            for ver, priv in zip(self.testnet_vers, self.testnet_privs):
                digest = bytearray(urandom(32))
                r, s, order = priv.raw_sign(digest)
                self.assertTrue(s in range(1, PrivateKey.highest_s))
                self.assertTrue(ver.verify_digest((r, s), digest, sigdecode=lambda sig, _: (sig[0], sig[1])))

    def test_der_sig_success(self):
        from os import urandom
        import struct

        for _ in range(10):

            for priv in self.mainnet_privs:
                digest = bytearray(urandom(32))
                sig = priv.sign(digest)
                length_r = sig[3]
                length_s = sig[5 + length_r]
                s = int.from_bytes(bytearray(struct.unpack(str(length_s) + 'B', sig[6 + length_r:6 + length_r + length_s])), 'big')
                self.assertTrue(s in range(1, PrivateKey.highest_s))

            for priv in self.testnet_privs:
                digest = bytearray(urandom(32))
                sig = priv.sign(digest)
                length_r = sig[3]
                length_s = sig[5 + length_r]
                s = int.from_bytes(bytearray(struct.unpack(str(length_s) + 'B', sig[6 + length_r:6 + length_r + length_s])), 'big')
                self.assertTrue(s in range(1, PrivateKey.highest_s))


    def test_derivation_success(self):
        for pub, priv in keys:
            try:
                net = BitcoinMainnet
                pu = ExtendedPublicKey.decode(net, pub).key
            except ValueError:
                net = BitcoinTestnet
                pu = ExtendedPublicKey.decode(net, pub).key

            pr = ExtendedPrivateKey.decode(net, priv).key
            self.assertTrue(pr.pub() == pu)
            self.assertTrue(pr.pub().hexlify() == pu.hexlify())


class TestP2pk(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pubk = '0384478d41e71dc6c3f9edde0f928a47d1b724c05984ebfb4e7d0422e80abe95ff'

    def test_success(self):
        script = P2pkScript(PublicKey.unhexlify(self.pubk))
        self.assertTrue(script.decompile() == '{} OP_CHECKSIG'.format(self.pubk))
        self.assertTrue(script.hexlify() == '21{}ac'.format(self.pubk))

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('21{}acac'.format(self.pubk))

    def test_matching_fail_long_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('22{}11ac'.format(self.pubk))

    def test_matching_fail_short_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('20{}ac'.format(self.pubk[:-2]))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('21{}ad'.format(self.pubk))


class TestP2pkh(unittest.TestCase):

    def setUp(self):
        self.pubk = PublicKey.unhexlify('0384478d41e71dc6c3f9edde0f928a47d1b724c05984ebfb4e7d0422e80abe95ff')
        self.pubkh = self.pubk.hash()
        self.address = Address.from_string(BitcoinTestnet, 'mquvJWnJJwTDUcdneQkUbrfN2wm9uiXd1p')

    def test_success_pubk(self):
        script = P2pkhScript(self.pubk)
        self.assertTrue(script.decompile() ==
                        'OP_DUP OP_HASH160 {} OP_EQUALVERIFY OP_CHECKSIG'.format(hexlify(self.pubkh).decode()))
        self.assertTrue(script.hexlify() == '76a914{}88ac'.format(hexlify(self.pubkh).decode()))

    def test_success_hash(self):
        script = P2pkhScript(self.pubkh)
        self.assertTrue(script.decompile() ==
                        'OP_DUP OP_HASH160 {} OP_EQUALVERIFY OP_CHECKSIG'.format(hexlify(self.pubkh).decode()))
        self.assertTrue(script.hexlify() == '76a914{}88ac'.format(hexlify(self.pubkh).decode()))

    def test_success_address(self):
        script = P2pkhScript(self.address)
        self.assertTrue(script.decompile() ==
                        'OP_DUP OP_HASH160 {} OP_EQUALVERIFY OP_CHECKSIG'.format(hexlify(self.pubkh).decode()))
        self.assertTrue(script.hexlify() == '76a914{}88ac'.format(hexlify(self.pubkh).decode()))

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('76a914{}88acac'.format(hexlify(self.pubkh).decode()))

    def test_matching_fail_long_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('79a915{}1188ac'.format(hexlify(self.pubkh).decode()))

    def test_matching_fail_short_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('79a913{}88ac'.format(hexlify(self.pubkh[:-1]).decode()))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            P2pkScript.unhexlify('79a914{}88ad'.format(hexlify(self.pubkh).decode()))


class TestPwpkhScript(unittest.TestCase):

    def setUp(self):
        self.pubk = PublicKey.unhexlify('0384478d41e71dc6c3f9edde0f928a47d1b724c05984ebfb4e7d0422e80abe95ff')
        self.pubkh = self.pubk.hash()

    def test_success_pubk(self):
        script = P2wpkhV0Script(self.pubk)
        self.assertTrue(script.decompile() ==
                        'OP_0 {}'.format(hexlify(self.pubkh).decode()))
        self.assertTrue(script.hexlify() == '0014{}'.format(hexlify(self.pubkh).decode()))

    def test_success_hash(self):
        script = P2wpkhV0Script(self.pubkh)
        self.assertTrue(script.decompile() ==
                        'OP_0 {}'.format(hexlify(self.pubkh).decode()))
        self.assertTrue(script.hexlify() == '0014{}'.format(hexlify(self.pubkh).decode()))

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wpkhV0Script.unhexlify('0014{}aa'.format(hexlify(self.pubkh).decode()))

    def test_matching_fail_long_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wpkhV0Script.unhexlify('0015{}aa'.format(hexlify(self.pubkh).decode()))

    def test_matching_fail_short_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wpkhV0Script.unhexlify('0013{}'.format(hexlify(self.pubkh[:-1]).decode()))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wpkhV0Script.unhexlify('0114{}'.format(hexlify(self.pubkh).decode()))


class TestP2sh(unittest.TestCase):

    def setUp(self):
        self.redeem_script = P2pkhScript(PublicKey.unhexlify('0384478d41e71dc6c3f9edde0f928a47d1b724c'
                                                             '05984ebfb4e7d0422e80abe95ff'))
        self.as_data = StackData.from_bytes(self.redeem_script.p2sh_hash())
        self.address = P2shAddress.from_script(BitcoinMainnet, self.redeem_script)

    def test_success_hash(self):
        script = P2shScript(self.redeem_script.p2sh_hash())
        self.assertTrue(script.decompile() ==
                        'OP_HASH160 {} OP_EQUAL'.format(self.as_data))
        self.assertTrue(script.hexlify() == 'a9{:02x}{}87'.format(len(self.as_data), self.as_data))

    def test_success_redeem(self):
        script = P2shScript(self.redeem_script)
        self.assertTrue(script.decompile() ==
                        'OP_HASH160 {} OP_EQUAL'.format(self.as_data))
        self.assertTrue(script.hexlify() == 'a9{:02x}{}87'.format(len(self.as_data), self.as_data))

    def test_success_addresses(self):

        for script_hex, address in p2sh.items():
            script = ScriptBuilder.identify(bytearray(unhexlify(script_hex)))
            try:
                net = BitcoinMainnet
                from_addr = P2shScript(Address.from_string(net, address))
            except CouldNotDecode:
                net = BitcoinTestnet
                from_addr = P2shScript(Address.from_string(net, address))

            from_script = P2shScript(script)
            self.assertTrue(str(from_addr.address(net)) == address)
            self.assertTrue(str(P2shAddress.from_script(net, script)) == address)
            self.assertTrue(str(from_script.address(net)) == address)

        script = P2shScript(self.address)
        self.assertTrue(script.decompile() ==
                        'OP_HASH160 {} OP_EQUAL'.format(self.as_data))
        self.assertTrue(script.hexlify() == 'a9{:02x}{}87'.format(len(self.as_data), self.as_data))

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript(Script('a9{:02x}{}87ac'.format(len(self.as_data), self.as_data)))

    def test_matching_fail_long_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript.unhexlify('a9{:02x}{}aa87'.format(len(self.as_data)+1, self.as_data))

    def test_matching_fail_short_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript.unhexlify('a9{:02x}{}87'.format(len(self.as_data)-1, str(self.as_data)[:-2]))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript.unhexlify('a9{:02x}{}88'.format(len(self.as_data), self.as_data))


class TestP2wsh(unittest.TestCase):

    def setUp(self):
        self.redeem_script = P2pkhScript(PublicKey.unhexlify('0384478d41e71dc6c3f9edde0f928a47d1b724c'
                                                             '05984ebfb4e7d0422e80abe95ff'))
        self.as_data = StackData.from_bytes(self.redeem_script.p2wsh_hash())

    def test_success_hash(self):
        script = P2wshV0Script(self.redeem_script.p2wsh_hash())
        self.assertTrue(script.decompile() ==
                        'OP_0 {}'.format(self.as_data))
        self.assertTrue(script.hexlify() == '00{:02x}{}'.format(len(self.as_data), self.as_data))

    def test_success_redeem(self):
        script = P2wshV0Script(self.redeem_script)
        self.assertTrue(script.decompile() ==
                        'OP_0 {}'.format(self.as_data))
        self.assertTrue(script.hexlify() == '00{:02x}{}'.format(len(self.as_data), self.as_data))

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wshV0Script.unhexlify('00{:02x}{}aa'.format(len(self.as_data), self.as_data))

    def test_matching_fail_long_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wshV0Script.unhexlify('00{:02x}{}aa'.format(len(self.as_data) + 1, self.as_data))

    def test_matching_fail_short_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wshV0Script.unhexlify('00{:02x}{}'.format(len(self.as_data) - 1, str(self.as_data)[:-2]))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            P2wshV0Script.unhexlify('01{:02x}{}'.format(len(self.as_data), self.as_data))


class TestNulldata(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = StackData.unhexlify('c97232cda052b9e9c312af692c310e4061dcf937910a335bdb96865a0fb9469b0622c6fa')

    def test_success(self):
        script = NulldataScript(self.data)
        self.assertTrue(script.decompile() ==
                        'OP_RETURN {}'.format(self.data))
        self.assertTrue(script.hexlify() == '6a{:02x}{}'.format(len(self.data), self.data))

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript.unhexlify('a9{:02x}{}aa'.format(len(self.data), self.data))

    def test_matching_fail_long_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript.unhexlify('a9{:02x}{}aa'.format(len(self.data)+1, self.data))

    def test_matching_fail_short_push(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript.unhexlify('a9{:02x}{}'.format(len(self.data)-1, str(self.data)[:-2]))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            P2shScript.unhexlify('a8{:02x}{}'.format(len(self.data), self.data))


class TestMultisig(unittest.TestCase):

    def setUp(self):
        self.m = 2
        self.pubkeys = [PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bfad09a89e211aeb926242"),
                        PublicKey.unhexlify("033e81519ecf373ea3a5c7e1c051b71a898fb3438c9550e274d980f147eb4d069d"),
                        PublicKey.unhexlify("036d568125a969dc78b963b494fa7ed5f20ee9c2f2fc2c57f86c5df63089f2ed3a")]
        self.n = 3
        self.hex_template = '{:02x}{}{:02x}ae'.format(self.m + 80,
                                                      ''.join('{:02x}{}'.format(len(pk), pk) for pk in self.pubkeys),
                                                      self.n + 80)

    def test_success(self):
        script = MultisigScript(self.m, *(self.pubkeys + [self.n]))
        expected = 'OP_{} {} OP_{} OP_CHECKMULTISIG'.format(self.m, ' '.join(str(pk) for pk in self.pubkeys), self.n)
        self.assertTrue(script.decompile() == expected)
        self.assertTrue(script.hexlify() == self.hex_template)

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            MultisigScript.unhexlify(self.hex_template+'aa')

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            MultisigScript.unhexlify(self.hex_template[:-2]+'aa')


class TestIf(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inner_script = MultisigScript(2,
                                           PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bfad09a89e2"
                                                               "11aeb926242"),
                                           PublicKey.unhexlify("033e81519ecf373ea3a5c7e1c051b71a898fb3438c9550e274d980f"
                                                               "147eb4d069d"),
                                           PublicKey.unhexlify("036d568125a969dc78b963b494fa7ed5f20ee9c2f2fc2c57f86c5df"
                                                               "63089f2ed3a"),
                                           3)


class TestIfElse(unittest.TestCase):

    def setUp(self):
        self.if_script = MultisigScript(2,
                                        PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bfad09a89e2"
                                                            "11aeb926242"),
                                        PublicKey.unhexlify("033e81519ecf373ea3a5c7e1c051b71a898fb3438c9550e274d980f"
                                                            "147eb4d069d"),
                                        PublicKey.unhexlify("036d568125a969dc78b963b494fa7ed5f20ee9c2f2fc2c57f86c5df"
                                                            "63089f2ed3a"),
                                        3)
        self.else_script = P2shScript(P2pkhScript(PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bf"
                                                                      "ad09a89e211aeb926242")))

    def test_success(self):
        script = IfElseScript(self.if_script, self.else_script)
        self.assertTrue(script.decompile() ==
                        'OP_IF {} OP_ELSE {} OP_ENDIF'.format(self.if_script.decompile(), self.else_script.decompile()))
        self.assertTrue(script.hexlify() == '63{}67{}68'.format(self.if_script.hexlify(), self.else_script.hexlify()))

    def test_matching_fail_leftover(self):
        with self.assertRaises(WrongScriptTypeException):
            IfElseScript.unhexlify('63{}67{}68aa'.format(self.if_script.hexlify(), self.else_script.hexlify()))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            IfElseScript.unhexlify('63{}aa{}68'.format(self.if_script.hexlify(), self.else_script.hexlify()))


class TestTimelock(unittest.TestCase):

    def setUp(self):
        self.locktime = 500000
        self.locked_script = P2shScript(P2pkhScript(PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bf"
                                                                        "ad09a89e211aeb926242")))

    def test_success(self):
        script = TimelockScript(Locktime(self.locktime), self.locked_script)
        self.assertTrue(script.decompile() ==
                        '{} OP_CHECKLOCKTIMEVERIFY OP_DROP {}'.format(Locktime(self.locktime).for_script(),
                                                                      self.locked_script.decompile()))
        self.assertTrue(script.hexlify() == '{}b175{}'.format(Locktime(self.locktime).for_script().hexlify(),
                                                              self.locked_script.hexlify()))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            TimelockScript.unhexlify('{}b1aa{}'.format(Locktime(self.locktime).for_script().hexlify(),
                                                       self.locked_script.hexlify()))


class TestRelativeTimelock(unittest.TestCase):

    def setUp(self):
        self.sequence = 500000
        self.locked_script = P2shScript(P2pkhScript(PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bf"
                                                                        "ad09a89e211aeb926242")))

    def test_success(self):
        script = RelativeTimelockScript(Sequence(self.sequence), self.locked_script, strict=True)
        self.assertTrue(script.decompile() ==
                        '{} OP_CHECKSEQUENCEVERIFY OP_DROP {}'.format(Sequence(self.sequence).for_script(),
                                                                      self.locked_script.decompile()))
        self.assertTrue(script.hexlify() == '{}b275{}'.format(Sequence(self.sequence).for_script().hexlify(),
                                                              self.locked_script.hexlify()))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            RelativeTimelockScript.unhexlify('{}b2aa{}'.format(Sequence(self.sequence).for_script().hexlify(),
                                                               self.locked_script.hexlify()))


class TestHashlock(unittest.TestCase):

    def setUp(self):
        self.hash = PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bfad09a89e211aeb926242").hash()
        self.locked_script = P2shScript(P2pkhScript(PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bf"
                                                                        "ad09a89e211aeb926242")))

    def test_success(self):
        script = Hashlock256Script(StackData.from_bytes(self.hash), self.locked_script)
        self.assertTrue(script.decompile() ==
                        'OP_HASH256 {} OP_EQUALVERIFY {}'.format(hexlify(self.hash).decode(),
                                                                 self.locked_script.decompile()))
        self.assertTrue(script.hexlify() == 'aa{:02x}{}88{}'.format(len(self.hash),
                                                                    hexlify(self.hash).decode(),
                                                                    self.locked_script.hexlify()))

    def test_matching_fail_wrong_op(self):
        with self.assertRaises(WrongScriptTypeException):
            Hashlock256Script.unhexlify('{:02x}{}a8aa{}'.format(len(self.hash),
                                                                hexlify(self.hash).decode(),
                                                                self.locked_script.hexlify()))


class TestBitcoinAddress(unittest.TestCase):

    def setUp(self):
        self.good_addresses = {('BitcoinMainnet', P2pkhAddress, '1KKKK6N21XKo48zWKuQKXdvSsCf95ibHFa',
                                b'\xc8\xe9\t\x96\xc7\xc6\x08\x0e\xe0b\x84`\x0chN\xd9\x04\xd1L\\'),
                               ('BitcoinTestnet', P2pkhAddress, 'mtH6FLMQNu2fFQ4mrb7UjEUjTAhUNCMFoi',
                                b'\x8b\xfas]\x98\xabb\x1e\xdbOw\xd7\xb7\xfe\nK\x1f\xfc\xc0$'),
                               ('BitcoinTestnet', P2pkhAddress, 'n3wEvhujG7SDcgeKCXZMrty5QqhQZ7f6jW',
                                b'\xf5\xea\xa2K\x82\xc8\x1f4L\x9a\x16\xa8\xfb\x84t\xe1\x10\xfd\xb1\xc3'),
                               ('BitcoinMainnet', P2shAddress, '3P14159f73E4gFr7JterCCQh9QjiTjiZrG',
                                b'\xe9\xc3\xdd\x0c\x07\xaa\xc7ay\xeb\xc7jlx\xd4\xd6|l\x16\n'),
                               ('BitcoinTestnet', P2shAddress, '2N6JFaB5rMtPwutovP6cirwBVxHuAVaHvMG',
                                b'\x8f,4\xa2<F\xe9\x80\xb7\x9e\x10\x9b\x11\xa2\xc8-9\x92\xeb\x95')}
        self.bad_addresses = {'vioqwV3F4YzpgnfyUukGVMB3Hv83ujehKCiGWyrYyx2Z7hiKQy7SWUV9KgfMdV9J',
                              'bc1a',
                              '3rE3tz',
                              '1KKKK6N21XKo48zWKuuKXdvSsCf95ibHFa'}

    def test_success(self):
        for net, addr_type, address, hashed_data in self.good_addresses:
            if net == 'BitcoinMainnet':
                net = BitcoinMainnet
            else:
                net = BitcoinTestnet

            from_string = Address.from_string(net, address)
            self.assertTrue(address == str(from_string))
            self.assertTrue(from_string.__class__ == addr_type)
            self.assertTrue(from_string.network == net)
            self.assertTrue(from_string.hash == hashed_data)

    def test_fail(self):
        for address in self.bad_addresses:
            with self.assertRaises(ValueError):
                Address.from_string(BitcoinMainnet, address)

    def test_conversions(self):
        for address, pkh in addresses:
            self.assertEqual(hexlify(Address.from_string(BitcoinMainnet, address).hash).decode(), pkh)
            self.assertEqual(str(P2pkhScript(bytearray(unhexlify(pkh))).address(BitcoinMainnet)), address)
            self.assertEqual(P2pkhScript(Address.from_string(BitcoinMainnet, address)).pubkeyhash,
                             bytearray(unhexlify(pkh)))
            self.assertEqual(P2pkhAddress(BitcoinMainnet, bytearray(unhexlify(pkh))).hash, bytearray(unhexlify(pkh)))


class TestPeercoinAddress(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.good_addresses = {
            ('PeercoinMainnet', P2pkhAddress, 'PAdonateFczhZuKLkKHozrcyMJW7Y6TKvw',),
            ('PeercoinTestnet', P2pkhAddress, 'mj46gUeZgeD9ufU7Fvz2dWqaX6Nswtbpba',),
            ('PeercoinTestnet', P2pkhAddress, 'n12h8P5LrVXozfhEQEqg8SFUmVKtphBetj',),
            ('PeercoinMainnet', P2shAddress, 'p92W3t7YkKfQEPDb7cG9jQ6iMh7cpKLvwK',),
        }
        self.bad_addresses = {
            'vioqwV3F4YzpgnfyUukGVMB3Hv83ujehKCiGWyrYyx2Z7hiKQy7SWUV9KgfMdV9J',
            'bc1a',
            '3rE3tz',
            '1KKKK6N21XKo48zWKuQKXdvSsCf95ibHFa',
        }

    def test_success(self):
        for net, addr_type, address in self.good_addresses:
            if net == 'PeercoinMainnet':
                net = PeercoinMainnet
            else:
                net = PeercoinTestnet

            from_string = Address.from_string(net, address)
            self.assertTrue(address == str(from_string))
            self.assertTrue(from_string.__class__ == addr_type)
            self.assertTrue(from_string.network == net)

    def test_fail(self):
        for address in self.bad_addresses:
            with self.assertRaises(ValueError):
                Address.from_string(PeercoinMainnet, address)


class TestStandardness(unittest.TestCase):

    def test_script_pubkey_success(self):
        # standard types
        self.assertTrue(eval(scripts['p2sh']['code']).is_standard())
        self.assertTrue(eval(scripts['p2pkh']['code']).is_standard())
        self.assertTrue(eval(scripts['p2pk']['code']).is_standard())
        self.assertTrue(eval(scripts['multisig']['code']).is_standard())
        self.assertTrue(eval(scripts['nulldata']['code']).is_standard())
        self.assertTrue(eval(scripts['p2wpkh']['code']).is_standard())
        self.assertTrue(eval(scripts['p2wsh']['code']).is_standard())

    def test_script_pubkey_fail(self):
        # nonstandard types
        self.assertFalse(eval(scripts['if_else_timelock']['code']).is_standard())
        self.assertFalse(eval(scripts['relativetimelock']['code']).is_standard())
        # n > 3
        self.assertFalse(
            MultisigScript(
                2,
                PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bfad09a89e211aeb926242"),
                PublicKey.unhexlify("033e81519ecf373ea3a5c7e1c051b71a898fb3438c9550e274d980f147eb4d069d"),
                PublicKey.unhexlify("036d568125a969dc78b963b494fa7ed5f20ee9c2f2fc2c57f86c5df63089f2ed3a"),
                PublicKey.unhexlify("036d568125a969dc78b963b494fa7ed5f20ee9c2f2fc2c57f86c5df63089f2ed3a"),
                4
            ).is_standard()
        )
        # m > n
        self.assertFalse(
            MultisigScript(
                4,
                PublicKey.unhexlify("02c08786d63f78bd0a6777ffe9c978cf5899756cfc32bfad09a89e211aeb926242"),
                PublicKey.unhexlify("033e81519ecf373ea3a5c7e1c051b71a898fb3438c9550e274d980f147eb4d069d"),
                PublicKey.unhexlify("036d568125a969dc78b963b494fa7ed5f20ee9c2f2fc2c57f86c5df63089f2ed3a"),
                3,
                strict=False).is_standard()
        )
        # >80-byte data
        self.assertFalse(NulldataScript(StackData.unhexlify(
            '444f4350524f4f463832bd18ceb0a7861f2a8198013047a3fb861261523c0fc4164abc044e517702444f4350524f4f463832bd18ce'
            'b0a7861f2a8198013047a3fb861261523c0fc4164abc044e51770211')).is_standard())

    def test_txin_success(self):
        # coinbase txin
        txin = Transaction.unhexlify('01000000010000000000000000000000000000000000000000000000000000000000000000'
                                     'ffffffff3d039920071c2f706f6f6c2e626974636f696e2e636f6d2f4249503130302f4238'
                                     '2f0a092f4542312f4144362f109c52640027ed852ba74c0741c3eb0100ffffffff01505266'
                                     '53000000001976a9143fa71ed2e38d431960f314e7e7aad476a5496b4c88ac00000000').ins[0]
        self.assertTrue(txin.is_standard())
        txin = Transaction.unhexlify('0100000001e4da173fbefe5e60ff63dfd38566ade407532294db655463b77a783f379ce6050000000'
                                     '06b483045022100af246c27890c2bc07a0b7450d3d82509702a44a4defdff766355240b114ee2ac02'
                                     '207bb67b468452fa1b325dd5583879f5c1412e0bb4dae1c2c96c7a408796ab76f1012102ab9e85755'
                                     '36a1e99604a158fc60fe2ebd1cb1839e919b4ca42b8d050cfad71b2ffffffff0100c2eb0b00000000'
                                     '1976a914df76c017354ac39bde796abe4294d31de8b5788a88ac00000000').ins[0]
        self.assertTrue(txin.is_standard())

        # actual redeem script (15 sigops: 14 OP_CHECKMULTISIGVERIFY + OP_CHECKSIG)
        script_sig = ScriptSig(bytearray(b'F0C\x02\x1f\x19\xeewU\xa5w\xd8G\x03\x8f;2K\x8b\xfb\xc4 \x006"\xe73\xa9'
                                         b'\x953\x88:,!\xc2\x16\x02 \x1asWN\x01\xb6fw\xc7\r\x93\x0c|\x19\x182W\x04'
                                         b'\xb0\xe6\xd0.\x92\xe6\xb0\xcdk\xc2\xf0n\r\x8e\x01(\x00c^\xafh!\x02\xae'
                                         b'p6\xec\xd9\xda\xc4\xef)\xff\x0b1\x03\xc1\xa8\xfb[\xa0\xbe=\xed\xf65\xd6'
                                         b'\xfa\xa6\x1f\x06K\xef\x07\xff\xac'))
        prev_script = P2shScript(bytearray(b'\xa9\x14\x00H\xee[\x92\xbc\xc9kx\xa0W\x1e\xdch\x14`\xe8B`\xa5\x87'))
        txin = TxIn('4a0884b0aa0c3d34a81ea747aca1368effd9359d573f79873d2d6b4045e49205',
                    9,
                    script_sig,
                    Sequence(0xffffffff))
        self.assertTrue(txin.is_standard(prev_script))

        # 14 sigops redeem script
        script_sig = ScriptSig(bytearray(b'F0C\x02\x1f\x19\xeewU\xa5w\xd8G\x03\x8f;2K\x8b\xfb\xc4 \x006"\xe73\xa9'
                                         b'\x953\x88:,!\xc2\x16\x02 \x1asWN\x01\xb6fw\xc7\r\x93\x0c|\x19\x182W\x04'
                                         b'\xb0\xe6\xd0.\x92\xe6\xb0\xcdk\xc2\xf0n\r\x8e\x01(\x00c^\xafh!\x02\xae'
                                         b'p6\xec\xd9\xda\xc4\xef)\xff\x0b1\x03\xc1\xa8\xfb[\xa0\xbe=\xed\xf65\xd6'
                                         b'\xfa\xa6\x1f\x06K\xef\x07\xff\xac'))
        prev_script = P2shScript(bytearray(b'\xa9\x14\x00H\xee[\x92\xbc\xc9kx\xa0W\x1e\xdch\x14`\xe8B`\xa5\x87'))
        txin = TxIn('4a0884b0aa0c3d34a81ea747aca1368effd9359d573f79873d2d6b4045e49205',
                    9,
                    script_sig,
                    Sequence(0xffffffff))
        self.assertTrue(txin.is_standard(prev_script))

    def test_txin_fail(self):
        # redeem script with 16 sigops (same as above + b'\xac')
        script_sig = ScriptSig(bytearray(b'F0C\x02\x1f\x19\xeewU\xa5w\xd8G\x03\x8f;2K\x8b\xfb\xc4 \x006"\xe73\xa9'
                                         b'\x953\x88:,!\xc2\x16\x02 \x1asWN\x01\xb6fw\xc7\r\x93\x0c|\x19\x182W\x04'
                                         b'\xb0\xe6\xd0.\x92\xe6\xb0\xcdk\xc2\xf0n\r\x8e\x01(\x00c^\xafh!\x02\xae'
                                         b'p6\xec\xd9\xda\xc4\xef)\xff\x0b1\x03\xc1\xa8\xfb[\xa0\xbe=\xed\xf65\xd6'
                                         b'\xfa\xa6\x1f\x06K\xef\x07\xff\xac' + b'\xac'))
        prev_script = P2shScript(bytearray(b'\xa9\x14\x00H\xee[\x92\xbc\xc9kx\xa0W\x1e\xdch\x14`\xe8B`\xa5\x87'))
        txin = TxIn('4a0884b0aa0c3d34a81ea747aca1368effd9359d573f79873d2d6b4045e49205',
                    9,
                    script_sig,
                    Sequence(0xffffffff))
        self.assertFalse(txin.is_standard(prev_script))

        # nonstandard prev_script
        script_sig = ScriptSig(bytearray(b'F0C\x02\x1f\x19\xeewU\xa5w\xd8G\x03\x8f;2K\x8b\xfb\xc4 \x006"\xe73\xa9'
                                         b'\x953\x88:,!\xc2\x16\x02 \x1asWN\x01\xb6fw\xc7\r\x93\x0c|\x19\x182W\x04'
                                         b'\xb0\xe6\xd0.\x92\xe6\xb0\xcdk\xc2\xf0n\r\x8e\x01(\x00c^\xafh!\x02\xae'
                                         b'p6\xec\xd9\xda\xc4\xef)\xff\x0b1\x03\xc1\xa8\xfb[\xa0\xbe=\xed\xf65\xd6'
                                         b'\xfa\xa6\x1f\x06K\xef\x07\xff\xac'))
        prev_script = UnknownScript('\xac')
        txin = TxIn('4a0884b0aa0c3d34a81ea747aca1368effd9359d573f79873d2d6b4045e49205',
                    9,
                    script_sig,
                    Sequence(0xffffffff))
        self.assertFalse(txin.is_standard(prev_script))

    def test_tx_success(self):
        pass

    def test_tx_fail(self):
        pass

    def test_witness_success(self):
        pass

    def test_witness_fail(self):
        pass


class TestSegwitSigs(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = segsig

    def test_hash_prevouts(self):
        for tx in self.data:
            unsigned = SegWitTransaction.unhexlify(tx['unsigned_tx'])
            self.assertEqual(unsigned._hash_prevouts(), bytearray(unhexlify(tx['hash_prevouts'])))

    def test_hash_sequence(self):
        for tx in self.data:
            unsigned = SegWitTransaction.unhexlify(tx['unsigned_tx'])
            self.assertEqual(unsigned._hash_sequence(), bytearray(unhexlify(tx['hash_sequence'])))

    def test_hash_outputs(self):
        for tx in self.data:
            unsigned = SegWitTransaction.unhexlify(tx['unsigned_tx'])
            self.assertEqual(unsigned._hash_outputs(), bytearray(unhexlify(tx['hash_outputs'])))

    def test_digest_preimage(self):
        for tx in self.data:
            unsigned = SegWitTransaction.unhexlify(tx['unsigned_tx'])
            for j, txin in enumerate(tx['txins']):
                if 'digest_preimage' in txin:
                    self.assertEqual(unsigned._get_segwit_digest_preimage(j,
                                                                          ScriptBuilder.identify(txin['prev_script']),
                                                                          txin['prev_amount']).hexlify(),
                                     txin['digest_preimage'])

    def test_digest(self):
        for tx in self.data:
            unsigned = SegWitTransaction.unhexlify(tx['unsigned_tx'])
            for j, txin in enumerate(tx['txins']):
                if 'digest' in txin:
                    self.assertEqual(hexlify(unsigned.get_segwit_digest(j,
                                                                        ScriptBuilder.identify(txin['prev_script']),
                                                                        txin['prev_amount'])).decode(),
                                     txin['digest'])

    def test_sig(self):
        for tx in self.data:
            if 'signature' in tx:
                for txin in tx['txins']:
                    if 'privk' in txin and 'digest' in txin:
                        self.assertEqual(hexlify(PrivateKey.unhexlify(txin['privk']).sign(bytearray(unhexlify(txin['digest'])),
                                                                                          deterministic=True)).decode(),
                                         tx['signature'])


class TestStrictMode(unittest.TestCase):

    def test_multisig_parsing_non_strict(self):
        # 1 of 2 with one valid and one invalid public key
        script_hex = ('5121037953dbf08030f67352134992643d033417eaa6fcfb770c038f364ff40d7615882100e8f87dd9d24c3a2f1'
                      '02a5a8276c4a8f58176c961dada423b61063a312b7c270e52ae')
        script = ScriptBuilder.identify(script_hex)
        self.assertTrue(isinstance(script, MultisigScript))
        self.assertEqual(script.type, 'multisig')
        self.assertTrue(isinstance(script.pubkeys[0], PublicKey))
        self.assertTrue(isinstance(script.pubkeys[1], StackData))

        # 1 of 2 with two invalid public keys
        script_hex = ('512100e8f87dd9d24c3a2f102a5a8276c4a8f58176c961dada423b61063a312b7c270e2100e8f87dd9d24c3a2f1'
                      '02a5a8276c4a8f58176c961dada423b61063a312b7c270e52ae')
        script = ScriptBuilder.identify(script_hex, strict=False)
        self.assertTrue(isinstance(script, MultisigScript))
        self.assertEqual(script.type, 'multisig')
        self.assertTrue(isinstance(script.pubkeys[0], StackData))
        self.assertTrue(isinstance(script.pubkeys[1], StackData))

        # 2 of 2 with one valid and one invalid public key
        script_hex = ('5221037953dbf08030f67352134992643d033417eaa6fcfb770c038f364ff40d7615882100e8f87dd9d24c3a2f1'
                      '02a5a8276c4a8f58176c961dada423b61063a312b7c270e52ae')
        script = ScriptBuilder.identify(script_hex, strict=False)
        self.assertTrue(isinstance(script, MultisigScript))
        self.assertEqual(script.type, 'multisig')
        self.assertTrue(isinstance(script.pubkeys[0], PublicKey))
        self.assertTrue(isinstance(script.pubkeys[1], StackData))

    def test_multisig_parsing_strict(self):
        # 1 of 2 with one valid and one invalid public key
        script_hex = ('5121037953dbf08030f67352134992643d033417eaa6fcfb770c038f364ff40d7615882100e8f87dd9d24c3a2f1'
                      '02a5a8276c4a8f58176c961dada423b61063a312b7c270e52ae')
        script = ScriptBuilder.identify(script_hex)
        self.assertTrue(isinstance(script, MultisigScript))
        self.assertEqual(script.type, 'multisig')
        self.assertTrue(isinstance(script.pubkeys[0], PublicKey))
        self.assertTrue(isinstance(script.pubkeys[1], StackData))

        # 1 of 2 with two invalid public keys
        script_hex = ('512100e8f87dd9d24c3a2f102a5a8276c4a8f58176c961dada423b61063a312b7c270e2100e8f87dd9d24c3a2f1'
                      '02a5a8276c4a8f58176c961dada423b61063a312b7c270e52ae')
        script = Script.unhexlify(script_hex)
        with self.assertRaises(WrongPubKeyFormat):
            MultisigScript(script)

        # 2 of 2 with one valid and one invalid public key
        script_hex = ('5221037953dbf08030f67352134992643d033417eaa6fcfb770c038f364ff40d7615882100e8f87dd9d24c3a2f1'
                      '02a5a8276c4a8f58176c961dada423b61063a312b7c270e52ae')
        script = Script.unhexlify(script_hex)
        with self.assertRaises(WrongPubKeyFormat):
            MultisigScript(script)

    def test_multisig_creation_strict(self):
        with self.assertRaises(WrongPubKeyFormat):
            MultisigScript(1, StackData.unhexlify('00'*33), StackData.unhexlify('00'*33), 2)

        script = MultisigScript(1,
                                StackData.unhexlify('00'*33),
                                PublicKey.unhexlify('0384478d41e71dc6c3f9edde0f928a47d1b724c05984ebfb4e7d0422e80abe95ff'),
                                2)
        self.assertTrue(isinstance(script, MultisigScript))
        self.assertEqual(script.type, 'multisig')
        self.assertTrue(isinstance(script.pubkeys[0], StackData))
        self.assertTrue(isinstance(script.pubkeys[1], PublicKey))

    def test_multisig_creation_non_strict(self):
        script = MultisigScript(
            1,
            StackData.unhexlify('00'*33),
            StackData.unhexlify('00'*33),
            2,
            strict=False,
        )
        self.assertTrue(isinstance(script, MultisigScript))
        self.assertEqual(script.type, 'multisig')
        self.assertTrue(isinstance(script.pubkeys[0], StackData))
        self.assertTrue(isinstance(script.pubkeys[1], StackData))

        script = MultisigScript(
            1,
            StackData.unhexlify('00'*33),
            PublicKey.unhexlify('0384478d41e71dc6c3f9edde0f928a47d1b724c05984ebfb4e7d0422e80abe95ff'),
            2,
            strict=False,
        )
        self.assertTrue(isinstance(script, MultisigScript))
        self.assertEqual(script.type, 'multisig')
        self.assertTrue(isinstance(script.pubkeys[0], StackData))
        self.assertTrue(isinstance(script.pubkeys[1], PublicKey))

        with self.assertRaises(WrongScriptTypeException):
            MultisigScript(
                1,
                StackData.unhexlify('00'*2),
                StackData.unhexlify('00'*33),
                2,
                strict=False,
            )

    def test_p2pk_parsing_non_strict(self):
        script_hex = '2100e8f87dd9d24c3a2f102a5a8276c4a8f58176c961dada423b61063a312b7c270eac'
        script = ScriptBuilder.identify(script_hex, strict=False)
        self.assertTrue(isinstance(script, P2pkScript))
        self.assertEqual(script.type, 'p2pk')
        self.assertTrue(isinstance(script.pubkey, StackData))

    def test_p2pk_parsing_strict(self):
        script_hex = '2100e8f87dd9d24c3a2f102a5a8276c4a8f58176c961dada423b61063a312b7c270eac'
        script = Script.unhexlify(script_hex)
        with self.assertRaises(WrongPubKeyFormat):
            P2pkScript(script)

    def test_p2pk_creation_strict(self):
        with self.assertRaises(TypeError):
            P2pkScript(StackData.unhexlify('00'*33))

    def test_p2pk_creation_non_strict(self):
        script = P2pkScript(StackData.unhexlify('00'*33), strict=False)
        self.assertTrue(isinstance(script, P2pkScript))
        self.assertEqual(script.type, 'p2pk')
        self.assertTrue(isinstance(script.pubkey, StackData))

        with self.assertRaises(WrongScriptTypeException):
            P2pkScript(StackData.unhexlify('00'*30), strict=False)


if __name__ == '__main__':
    unittest.main()
